#!/usr/bin/python

from SmartMeshSDK.ApiDefinition import ApiDefinition
from SmartMeshSDK.ApiDefinition.ByteArraySerializer import ByteArraySerializer

class IpMgrDefinition(ApiDefinition.ApiDefinition):
    '''
    \ingroup ApiDefinition
    
    \brief API definition for the IP manager.
   
    \note This class inherits from ApiDefinition. It redefines the attributes of
          its parents class, but inherits the methods.
    '''
    
    STRING    = ApiDefinition.FieldFormats.STRING
    BOOL      = ApiDefinition.FieldFormats.BOOL
    INT       = ApiDefinition.FieldFormats.INT
    INTS      = ApiDefinition.FieldFormats.INTS
    HEXDATA   = ApiDefinition.FieldFormats.HEXDATA
    ARRAY     = ApiDefinition.FieldFormats.ARRAY
    RC        = ApiDefinition.ApiDefinition.RC
    SUBID1    = ApiDefinition.ApiDefinition.SUBID1
    SUBID2    = ApiDefinition.ApiDefinition.SUBID2
    RC_OK     = ApiDefinition.ApiDefinition.RC_OK
    
    OPTIONAL  = [
        'pkLen_1',
        'delay_1',
        'pkLen_2',
        'delay_2',
        'pkLen_3',
        'delay_3',
        'pkLen_4',
        'delay_4',
        'pkLen_5',
        'delay_5',
        'pkLen_6',
        'delay_6',
        'pkLen_7',
        'delay_7',
        'pkLen_8',
        'delay_8',
        'pkLen_9',
        'delay_9',
        'pkLen_10',
        'delay_10',
        'stationId',
    ]
    
    def __init__(self, array2scalar = True):
        ApiDefinition.ApiDefinition.__init__(self, array2scalar)
        self.serializer = ByteArraySerializer(self)
    
    def default_serializer(self,commandArray,fieldsToFill):
        '''
        \brief IpMgrDefinition-specific implementation of default serializer
       
        \param commandArray   An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
        \param fieldsToFill   The list of fields to send, organized as a
                              dictionary of the form
                              <tt>fieldname:fieldvalue</tt>.
       
        \returns id,byteString where id is the command ID and byteArray an array
                               of bytes
        '''
        return self.serializer.serialize(commandArray,fieldsToFill)
    
    def deserialize(self,type,cmdId,byteArray):
        '''
        \brief IpMgrDefinition-specific implementation of deserializer
        '''
        return self.serializer.deserialize(type,cmdId,byteArray)
    
    # We redefine this attribute inherited from ApiDefinition. See
    # ApiDefinition for a full description of the structure of this field.
    fieldOptions = {
        # 'notificationTypes' : [
        #     [ 1,   'event',                 'Event notification'],
        #     [ 2,   'log',                   'Log notification'],
        #     [ 4,   'data',                  'Data payload notification'],
        #     [ 5,   'ipData',                '6LoWPAN packet notification'],
        #     [ 6,   'healthReport',          'Health report notification'],
        # ],
        # 'eventTypes' : [
        #     [ 0,   'moteReset',             'A mote reset'],
        #     [ 1,   'networkReset',          'The network was reset'],
        #     [ 2,   'commandFinish ',        'A command has completed execution'],
        #     [ 3,   'moteJoin',              'A mote joined the network'],
        #     [ 4,   'moteOperational',       'A new mote was configured and is now operational'],
        #     [ 5,   'moteLost',              'A mote is no longer communicating in the network'],
        #     [ 6,   'netTime',               'Contains the network uptime (in response to a getTime command)'],
        #     [ 7,   'pingResponse',          'A reply was received from a mote ping'],
        #     [10,   'pathCreate',            'A path was created'],
        #     [11,   'pathDelete',            'A path was deleted'],
        #     [12,   'packetSent',            'A packet was sent'],
        #     [13,   'moteCreate',            'A mote was created'],
        #     [14,   'moteDelete',            'A mote was deleted'],
        # ],
        RC : [
            [ 0,   'RC_OK',                 'The application layer has processed the command correctly'],
            [ 1,   'RC_INVALID_COMMAND',    'Invalid command'],
            [ 2,   'RC_INVALID_ARGUMENT',   'Invalid argument'],
            [11,   'RC_END_OF_LIST',        'End of list is returned when an iteration reaches the end of the list of objects'],
            [12,   'RC_NO_RESOURCES',       'Reached maximum number of items'],
            [13,   'RC_IN_PROGRESS',        'Operation is in progress'],
            [14,   'RC_NACK',               'Negative acknowledgment'],
            [15,   'RC_WRITE_FAIL',         'Flash write failed'],
            [16,   'RC_VALIDATION_ERROR',   'Parameter validation error'],
            [17,   'RC_INV_STATE',          'Object has inappropriate state'],
            [18,   'RC_NOT_FOUND',          'Object is not found'],
            [19,   'RC_UNSUPPORTED',        'The operation is not supported'],
        ],
        'frameProfile' : [
            [ 1,   'Profile_01',            'Fast network build, medium speed network operation'],
        ],
        'advertisementState' : [
            [ 0,   'on',                    'Advertisement is on'],
            [ 1,   'off',                   'Advertisement is off or slow*'],
        ],
        'downstreamFrameMode' : [
            [ 0,   'normal',                'Normal downstream bandwidth'],
            [ 1,   'fast',                  'Fast downstream bandwidth'],
        ],
        'networkState' : [
            [ 0,   'operational',           'Network is operating normally'],
            [ 1,   'radiotest',             'Manager is in radiotest mode'],
            [ 2,   'notStarted',            'Waiting for startNetwork API command'],
            [ 3,   'errorStartup',          'Unexpected error occurred at startup'],
            [ 4,   'errorConfig',           'Invalid or not licensed configuration found at startup'],
            [ 5,   'errorLicense',          'Invalid license file found at startup'],
        ],
        'moteState' : [
            [ 0,   'lost',                  'Mote is not currently part of the network'],
            [ 1,   'negotiating',           'Mote is in the process of joining the network'],
            [ 4,   'operational',           'Mote is operational'],
        ],
        'resetType' : [
            [ 0,   'resetSystem',           'Reset the system'],
            [ 2,   'resetMote',             'Reset the mote'],
        ],
        'backboneFrameMode' : [
            [ 0,   'off',                   'Backbone frame is off'],
            [ 1,   'upstream',              'Backbone frame is activated for upstream frames'],
            [ 2,   'bidirectional',         'Backbone frame is activated for both upstream and downstream frames'],
        ],
        'pathFilter' : [
            [ 0,   'all',                   'All paths'],
            [ 1,   'upstream',              'Upstream paths'],
        ],
        'pathDirection' : [
            [ 0,   'none',                  'No path'],
            [ 1,   'unused',                'Path is not used'],
            [ 2,   'upstream',              'Upstream path'],
            [ 3,   'downstream',            'Downstream path'],
        ],
        'packetPriority': [
            [0,    'Low',                   'Default packet priority'],
            [1,    'Medium',                'Higher packet priority'],
            [2,    'High',                  'Highest packet priority'],
        ],
        'commandFinishedResult': [
            [0,    'OK',                    'Command completed successfully'],
            [1,    'nack',                  'Command not acknowledged'],
            [2,    'commandTimeout',        'Command timed out'],
        ],
        'ccaMode': [
            [0,    'off',                   'CCA disabled'],
            [1,    'energy',                'Energy detect'],
            [2,    'carrier',               'Carrier detect'],
            [3,    'both',                  'Energy detect and Carrier detect'],
        ],
        'radiotestType': [
            [0,    'packet',                'Transmit packets'],
            [1,    'cm',                    'Continuous modulation'],
            [2,    'cw',                    'Continuous wave'],
            [3,    'pkcca',                 'Packet test with clear channel assessment (CCA) enabled'],
        ],
        'seqSize': [
            [0,    '0',                     ''],
            [1,    '1',                     ''],
            [2,    '2',                     ''],
            [3,    '3',                     ''],
            [4,    '4',                     ''],
            [5,    '5',                     ''],
            [6,    '6',                     ''],
            [7,    '7',                     ''],
            [8,    '8',                     ''],
            [9,    '9',                     ''],
            [10,   '10',                    ''],
        ],
        'joinFailedReason': [
            [0,    'counter',               'The join packet reused an already used join counter'],
            [1,    'notOnACL',               'The mote is not listed on the ACL'],
            [2,    'authentication',        'The join request could not be decrypted. Generally, this means the request was encrypted with a join key that did not match the key in the ACL.'],
            [3,    'unexpected',            'An unexpected error occurred while processing the join request'],
        ],
        
        
        #==================== misc ============================================
        'successCode' : [
            [ 0,   'success',               ''],
            [ 1,   'unsupported_version',   ''],
            [ 2,   'invalid_mode',          ''],
        ],
        'mode' : [
            [ 0,   'legacy',                ''],
        ],
        'userRole' : [
            [ 0,   'viewer',                'Viewer-role user has read-only access to non-sensitive network information'],
            [ 1,   'user',                  'User-role user has read-write privileges'],
        ],
        'protocolVersion' : 4,
    }
    
    # We redefine this attribute inherited from ApiDefinition. See
    # ApiDefinition for a full description of the structure of this field.
    commands = [
        # command 'hello' (commandID 1) is handled by the Hdlc module
        # command 'hello_response' (commandID 2) is handled by the Hdlc module
        {
            'id'         :  1,
            'name'       :  'mux_hello',
            'description':  'Sent by the manager to initiate a new session with a client.',
            'request'    : [
                ['version',                 INT,      1,   None],
                ['secret',                  HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['version',             INT,      1,   None],
                ],
            },
        },
        {
            'id'         :  1,
            'name'       :  'hello',
            'description':  '',
            'request'    : [
                ['version',                 INT,      1,   None],
                ['cliSeqNo',                INT,      1,   None],
                ['mode',                    INT,      1,   True],
            ],
        },
        {
            'id'         : 2,
            'name'       : 'hello_response',
            'description': '',
            'request'    : [], # there is no request
            'response'   : {
                'FIELDS':  [
                    ['successCode',         INT,      1,   True],
                    ['version',             INT,      1,   None],
                    ['mgrSeqNo',            INT,      1,   None],
                    ['cliSeqNo',            INT,      1,   None],
                    ['mode',                INT,      1,   True],
                ],
            },
        },
        {
            'id'         : 21,
            'name'       : 'reset',
            'description': 'The reset command is used to reset various objects. The command argument is an object type, and if the object is a mote the MAC address must be specified (otherwise that argument is ignored).',
            'request'    : [
                ['type',                    INT,      1,   'resetType'],
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['macAddress',          HEXDATA,  8,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Mote with specified MAC address is not found',
               'RC_INV_STATE'          : 'Mote is not in operational state',
               'RC_NACK'               : 'User commands queue is full (applies to mote reset)',
               'RC_INVALID_ARGUMENT'   : 'Invalid reset type value',
            },
        },
        {
            'id'         : 22,
            'name'       : 'subscribe',
            'description': "The subscribe command indicates that the manager should send the external application the specified notifications. It contains two filter fields:\n\n- filter is a bitmask of flags indicating the types of notifications that the client wants to receive\n- unackFilter allows the client to select which of the notifications selected in filter should be sent acknowledged. If a notification is sent as 'acknowledged', the subsequent notification packets will be queued while waiting for response.\n\nEach subscription request overwrites the previous one. If an application is subscribed to data and then decides he also wants events he should send a subscribe command with both the data and event flags set. To clear all subscriptions, the client should send a subscribe command with the filter set to zero. When a session is initiated between the manager and a client, the subscription filter is initialized to zero.\n\nThe subscribe bitmap uses the values of the notification type enumeration. Some values are unused to provide backwards compatibility with earlier APIs.",
            'request' : [
                ['filter',                  HEXDATA,  4,   None],
                ['unackFilter',             HEXDATA,  4,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_INVALID_ARGUMENT'   : 'Invalid subscription filter value',
            },
        },
        {
            'id'         : 23,
            'name'       : 'getTime',
            'description': 'The getTime command returns the current manager UTC time and current absolute slot number (ASN). The time values returned by this command are delayed by queuing and transfer time over the serial connection. For additional precision, an external application should trigger the networkTime notification using the Time Pin.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['uptime',              INT,      4,   None],
                    ['utcSecs',             INT,      8,   None],
                    ['utcUsecs',            INT,      4,   None],
                    ['asn',                 HEXDATA,  5,   None],
                    ['asnOffset',           INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 26,
            'name'       : 'setNetworkConfig',
            'description': 'The setNetworkConfig command changes network configuration parameters. The response code indicates whether the changes were successfully applied. This change is persistent.\n\nGenerally, changes to network configuration will take effect when the manager reboots. Exceptions are detailed below:\n\n- Max Motes: The new maxMotes value is used as soon as new motes try to join the network, but motes are not removed from the network if the value is set to a number lower than numMotes.\n- Base Bandwidth: Changing baseBandwidth while the network is running does not reallocate bandwidth to Operational motes.',
            'request'    :  [
                ['networkId',               INT,      2,   None],
                ['apTxPower',               INTS,     1,   None],
                ['frameProfile',            INT,      1,   True],
                ['maxMotes',                INT,      2,   None],
                ['baseBandwidth',           INT,      2,   None],
                ['downFrameMultVal',        INT,      1,   None],
                ['numParents',              INT,      1,   None],
                ['ccaMode',                 INT,      1,   True],
                ['channelList',             INT,      2,   None],
                ['autoStartNetwork',        BOOL,     1,   None],
                ['locMode',                 INT,      1,   None],
                ['bbMode',                  INT,      1,   'backboneFrameMode'],
                ['bbSize',                  INT,      1,   None],
                ['isRadioTest',             INT,      1,   None],
                ['bwMult',                  INT,      2,   None],
                ['oneChannel',              INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_INVALID_ARGUMENT'   : 'Validation of the network parameters failed',
               'RC_WRITE_FAIL'         : 'Flash write error, cannot save new settings',
            },
        },
        {
            'id'         : 31,
            'name'       : 'clearStatistics',
            'description': 'The clearStatistics command clears the accumulated network statistics. The command does not clear path quality or mote statistics.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 33,
            'name'       : 'exchangeMoteJoinKey',
            'description': 'The exchangeMoteJoinKey command triggers the manager to send a new join key to the specified mote and update the manager\'s ACL entry for the mote. The response contains a callbackId. A commandFinished event notification with this callbackId will be sent when the operation is complete. This change is persistent.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['key',                     HEXDATA,  16,  None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Mote with specified MAC address is not found',
               'RC_INV_STATE'          : 'Mote is not in operational state',
               'RC_NACK'               : 'User commands queue is full',
               'RC_WRITE_FAIL'         : 'Flash write error, can\'t save new settings',
            },
        },
        {
            'id'         : 34,
            'name'       : 'exchangeNetworkId',
            'description': 'The exchangeNetworkId command triggers the manager to distribute a new network ID to all the motes in the network. A callbackId is returned in the response. A commandFinished notification with this callbackId will be sent when the operation is complete. This change is persistent.',
            'request'    : [
                ['id',                      INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
                'RC_OK'                : 'Command received',
                'RC_IN_PROGRESS'       : 'A command is still pending. Wait until a commandFinished notification is received for the previous command before retrying.',
                'RC_NACK'              : 'User commands queue is full',
                'RC_WRITE_FAIL'        : 'Flash write error; cannot save new settings',
            },
        },
        {
            'id'         : 35,
            'name'       : 'radiotestTx',
            'description': 'The radiotestTx command allows the user to initiate a radio transmission test. It may only be executed if the manager has been booted up in radiotest mode (see setNetworkConfig command). Four types of transmission tests are supported:\n\n- Packet transmission\n- Continuous modulation (CM)\n- Continuous wave, i.e unmodulated signal (CW)\n- Packet transmission with clear channel assessment (CCA) enabled (Available in Manager > 1.3.x)\n\nIn a packet transmission test, the device generates a repeatCnt number of packet sequences. Each sequence consists of up to 10 packets with configurable size and delays. Each packet starts with a PHY preamble (5 bytes), followed by a PHY length field (1 byte), followed by data payload of up to 125 bytes, and finally a 2-byte 802.15.4 CRC at the end. Byte 0 of the payload contains stationId of the sender. Bytes 1 and 2 contain the packet number (in big-endian format) that increments with every packet transmitted. Bytes 3..N contain a counter (from 0..N-2) that increments with every byte inside payload. Transmissions occur on the set of channels defined by chanMask , selected in pseudo-random order.\n\nIn a continuous modulation test, the device generates continuous pseudo-random modulated signal, centered at the specified channel. The test is stopped by resetting the device.\n\nIn a continuous wave test, the device generates an unmodulated tone, centered at the specified channel. The test tone is stopped by resetting the device.\n\nIn a packet transmission with CCA test, the device is configured identically to that in the packet transmission test, however the device does a clear channel assessment before each transmission and aborts that packet if the channel is busy.\n\nThe station ID is a user selectable value. It is used in packet tests so that a receiver (see radiotestRx) can identify packets from this device in cases where there may be multiple tests running in the same radio space. This field is not used for CM or CW tests. (Available in Manager >= 1.3.0) Channel numbering is 0-15, corresponding to IEEE 2.4 GHz channels 11-26.',
            'request'    : [
                ['testType',                INT,      1,   'radiotestType'],
                ['chanMask',                HEXDATA,  2,   None],
                ['repeatCnt',               INT,      2,   None],
                ['txPower',                 INTS,     1,   None],
                ['seqSize',                 INT,      1,   'seqSize'],
                ['pkLen_1',                 INT,      1,   None],
                ['delay_1',                 INT,      2,   None],
                ['pkLen_2',                 INT,      1,   None],
                ['delay_2',                 INT,      2,   None],
                ['pkLen_3',                 INT,      1,   None],
                ['delay_3',                 INT,      2,   None],
                ['pkLen_4',                 INT,      1,   None],
                ['delay_4',                 INT,      2,   None],
                ['pkLen_5',                 INT,      1,   None],
                ['delay_5',                 INT,      2,   None],
                ['pkLen_6',                 INT,      1,   None],
                ['delay_6',                 INT,      2,   None],
                ['pkLen_7',                 INT,      1,   None],
                ['delay_7',                 INT,      2,   None],
                ['pkLen_8',                 INT,      1,   None],
                ['delay_8',                 INT,      2,   None],
                ['pkLen_9',                 INT,      1,   None],
                ['delay_9',                 INT,      2,   None],
                ['pkLen_10',                INT,      1,   None],
                ['delay_10',                INT,      2,   None],
                ['stationId',               INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_IN_PROGRESS'        : 'Radiotest is in progress',
               'RC_INVALID_ARGUMENT'   : 'Invalid "channel" or "txPower" value',
            },
            'serializer' : 'serialize_radiotestTx',
        },
        {
            'id'         : 37,
            'name'       : 'radiotestRx',
            'description': 'The radiotestRx command clears all previously collected statistics and initiates radio reception on the specified channel. It may only be executed if the manager has been booted up in radiotest mode (see setNetworkConfig command). During the test, the device keeps statistics about the number of packets received (with and without error). The test results may be retrieved using the getRadiotestStatistics command.\n\nThe station ID is a user selectable value. It must be set to match the station ID used by the transmitter. Station ID is used to isolate traffic if multiple tests are running in the same radio space.\n\n\n\nChannel numbering is 0-15, corresponding to IEEE 2.4 GHz channels 11-26.',
            'request'    : [
                ['mask',                    HEXDATA,  2,   None],
                ['duration',                INT,      2,   None],
                ['stationId',               INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_IN_PROGRESS'        : 'Radiotest is in progress',
               'RC_INVALID_ARGUMENT'   : 'Invalid mask value',
            },
        },
        {
            'id'         : 38,
            'name'       : 'getRadiotestStatistics',
            'description': 'This command retrieves statistics from a previously run radiotestRx command. It may only be executed if the manager has been booted up in radiotest mode (see setNetworkConfig command).',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['rxOk',                INT,      2,   None],
                    ['rxFail',              INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_IN_PROGRESS'        : 'Radiotest is in progress',
               'RC_INVALID_COMMAND'    : 'No radiotest was started',
            },
        },
        {
            'id'         : 39,
            'name'       : 'setACLEntry',
            'description': 'The setACLEntry command adds a new entry or updates an existing entry in the Access Control List (ACL). This change is persistent. The maximum number of entries is 1,200.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['joinKey',                 HEXDATA,  16,  None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NO_RESOURCES'       : 'ACL is full (when adding a new entry)',
               'RC_WRITE_FAIL'         : 'Flash write error, can\'t save new settings',
            },
        },
        {
            'id'         : 40,
            'name'       : 'getNextACLEntry',
            'description': 'The getNextACLEntry command returns information about next mote entry in the access control list (ACL). To begin a search (find the first mote in ACL), a zero MAC address (0000000000000000) should be sent. There is no mechanism for reading the ACL entry of a specific mote. This call is an iterator. If you call getNextACLEntry with mote A as the argument, your response is the ACL entry for mote B, where B is the next mote in the ACL.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['macAddress',          HEXDATA,  8,   None],
                    ['joinKey',             HEXDATA,  16,  None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_END_OF_LIST'        : 'End of ACL is reached',
               'RC_NOT_FOUND'          : 'No such mote in the ACL',
            },
        },
        {
            'id'         : 41,
            'name'       : 'deleteACLEntry',
            'description': 'The deleteACLEntry command deletes the specified mote from the access control list (ACL). If the macAddress parameter is set to all 0xFFs or all 0x00s, the entire ACL is cleared. This change is persistent.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Specified mote is not found in ACL',
               'RC_WRITE_FAIL'         : 'Flash write error, can\'t save new settings',
            },
        },
        {
            'id'         : 42,
            'name'       : 'pingMote',
            'description': 'The pingMote command sends a ping (echo request) to the mote specified by MAC address. A unique callbackId is generated and returned with the response. When the response is received from the mote, the manager generates a pingResponse notification with the measured round trip delay and several other parameters. The request is sent using unacknowledged transport, so the mote is not guaranteed to receive the request.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Specified mote not found',
               'RC_INV_STATE'          : 'Mote is not in operational state',
               'RC_NO_RESOURCES'       : 'User commands queue is full',
               'RC_IN_PROGRESS'        : 'Previous echo request command is still pending for specified mote',
            },
        },
        {
            'id'         : 43,
            'name'       : 'getLog',
            'description': 'The getLog command retrieves diagnostic logs from the manager or a mote specified by MAC address.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Specified mote not found',
               'RC_INV_STATE'          : 'Mote is not in operational state',
            },
        },
        {
            'id'         : 44,
            'name'       : 'sendData',
            'description': "The sendData command sends a packet to a mote in the network. The response contains a callbackId. When the manager injects the packet into the network, it will generate a packetSent notification. It is the responsibility of the customer's application layer at the mote to send a response. It is also the responsibility of the customer's application layer to timeout if no response is received at the manager if one is expected.\n\nThe sendData command should be used by applications that communicate directly with the manager. If end-to-end (application to mote) IP connectivity is required, the application should use the sendIP command. For a more comprehensive discussion of the distinction, see the SmartMesh IP Network User Guide.",
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['priority',                INT,      1,   'packetPriority'],
                ['srcPort',                 INT,      2,   None],
                ['dstPort',                 INT,      2,   None],
                ['options',                 INT,      1,   None],
                ['data',                    HEXDATA,  None,None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Specified mote is not found',
               'RC_INV_STATE'          : 'Mote is not in operational state',
               'RC_NACK'               : 'User commands queue is full or couldn\'t allocate memory buffer for payload',
               'RC_INVALID_ARGUMENT'   : 'Payload size exceeds maximum allowed value',
            },
        },
        {
            'id'         : 45,
            'name'       : 'startNetwork',
            'description': 'The startNetwork command tells the manager to allow the network to start forming (begin accepting join requests from devices). The external application must issue the startNetwork command if the autoStartNetwork flag is not set (see setNetworkConfig).\n\nThis command has been deprecated and should not be used in new designs.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_IN_PROGRESS'        : 'The network is already started',
            },
        },
        {
            'id'         : 46,
            'name'       : 'getSystemInfo',
            'description': 'The getSystemInfo command returns system-level information about the hardware and software versions.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['macAddress',          HEXDATA,  8,   None],
                    ['hwModel',             INT,      1,   None],
                    ['hwRev',               INT,      1,   None],
                    ['swMajor',             INT,      1,   None],
                    ['swMinor',             INT,      1,   None],
                    ['swPatch',             INT,      1,   None],
                    ['swBuild',             INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 47,
            'name'       : 'getMoteConfig',
            'description': 'The getMoteConfig command returns a single mote description as the response. The command takes two arguments, a MAC Address and a flag indicating whether the MAC Address refers to the requested mote or to the next mote in managers memory. This command may be used to iterate through all motes known by the manager by starting with the macAddress parameter set to 0 and next set to true, and then using the MAC Address of that response as the input to the next call.\n\nThe mote MAC address is used in all query commands, but space constraints require the neighbor health reports to use the Mote ID for identification. Therefore, both identifiers are present in the mote structure.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['next',                    BOOL,     1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['macAddress',          HEXDATA,  8,   None],
                    ['moteId',              INT,      2,   None],
                    ['isAP',                BOOL,     1,   None],
                    ['state',               INT,      1,   'moteState'],
                    ['reserved',            INT,      1,   None],
                    ['isRouting',           BOOL,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'The specified mote doesn\'t exist',
               'RC_END_OF_LIST'        : 'Last mote in the list has been reached (next = true)',
            },
        },
        {
            'id'         : 48,
            'name'       : 'getPathInfo',
            'description': 'The getPathInfo command returns parameters of requested path.',
            'request'    : [
                ['source',                  HEXDATA,  8,   None],
                ['dest',                    HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['source',              HEXDATA,  8,   None],
                    ['dest',                HEXDATA,  8,   None],
                    ['direction',           INT,      1,   'pathDirection'],
                    ['numLinks',            INT,      1,   None],
                    ['quality',             INT,      1,   None],
                    ['rssiSrcDest',         INTS,     1,   None],
                    ['rssiDestSrc',         INTS,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'A path between the specified motes doesn\'t exist',
            },
        },
        {
            'id'         : 49,
            'name'       : 'getNextPathInfo',
            'description': 'The getNextPathInfo command allows iteration across paths connected to a particular mote. The pathId parameter indicates the previous value in the iteration. Setting pathId to 0 returns the first path. A pathId can not be used as a unique identifier for a path. It is only valid when associated with a particular mote.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['filter',                  INT,      1,   'pathFilter'],
                ['pathId',                  INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['pathId',              INT,      2,   None],
                    ['source',              HEXDATA,  8,   None],
                    ['dest',                HEXDATA,  8,   None],
                    ['direction',           INT,      1,   'pathDirection'],
                    ['numLinks',            INT,      1,   None],
                    ['quality',             INT,      1,   None],
                    ['rssiSrcDest',         INTS,     1,   None],
                    ['rssiDestSrc',         INTS,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'The specified path ID does not exist',
               'RC_END_OF_LIST'        : 'The specified pathId in the request is the end of the list',
            },
        },
        {
            'id'         : 50,
            'name'       : 'setAdvertising',
            'description': '''The setAdvertising command tells the manager to activate, deactivate, or use slow advertising. The response is a callbackId. A commandFinished notification with the callbackId is generated when the command propagation is complete.

With motes prior to version 1.4.1, it is only possible to turn advertising ON or OFF. If building networks consisting primarily of motes 1.4.1 or later, power can be saved by setting advertising to "slow". Set the INI parameter advtimeout to a value (in ms) and set this command to 0.

For example, the default full advertising frequency is approximately once per 2 seconds. It is recommended to set advtimeout = 20000, which will result in an advertising every 20 seconds which will result in a 90% power savings in the cost of advertising.

It is dangerous to turn off advertising in the network. When advertising is off, new motes can not join and existing motes can not rejoin the network after a reset. Turning off advertising is primarily used to save power, or may be useful in for specific use cases where it is desirable to prevent motes from joining the network. In most cases, it is best to allow advertising to remain under the control of the manager.''',
            'request'    : [
                ['activate',                INT,      1,   'advertisementState'],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
                'RC_OK'                : 'Command successfully completed',
                'RC_IN_PROGRESS'       : 'A command is still pending. Wait until a commandFinished notification is received for the previous command before retrying.',
            },
        },
        {
            'id'         : 51,
            'name'       : 'setDownstreamFrameMode',
            'description': 'The setDownstreamFrameMode command tells the manager to shorten or extend the downstream slotframe. The base slotframe length will be multiplied by the downFrameMultVal for "normal" speed. For "fast" speed the downstream slotframe is the base length. Once this command is executed, the manager switches to manual mode and no longer changes slotframe size automatically. The response is a callbackId. A commandFinished notification with the callbackId is generated when the command propagation is complete.',
            'request'    : [
                ['frameMode',               INT,      1,   'downstreamFrameMode'],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_IN_PROGRESS'        : 'A command is still pending. Wait until a commandFinished notification is received for the previous command before retrying.',
               'RC_INVALID_ARGUMENT'   : 'The downFrameMultVal (as set by setNetworkConfig) is equal to 1, so changing the downstream frame mode would have no effect.',
            },
        },
        {
            'id'         : 53,
            'name'       : 'getManagerStatistics',
            'description': 'The getManagerStatistics command returns dynamic information and statistics about the manager API. The statistics counts are cleared together with all current statistics using clearStatistics.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['serTxCnt',            INT,      2,   None],
                    ['serRxCnt',            INT,      2,   None],
                    ['serRxCRCErr',         INT,      2,   None],
                    ['serRxOverruns',       INT,      2,   None],
                    ['apiEstabConn',        INT,      2,   None],
                    ['apiDroppedConn',      INT,      2,   None],
                    ['apiTxOk',             INT,      2,   None],
                    ['apiTxErr',            INT,      2,   None],
                    ['apiTxFail',           INT,      2,   None],
                    ['apiRxOk',             INT,      2,   None],
                    ['apiRxProtErr',        INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 54,
            'name'       : 'setTime',
            'description': 'This command has been deprecated, and should not be used in new designs. When the Manager restarts, it will start counting from 20:00:00 UTC July 2, 2002.\n\nThe setTime command sets the UTC time on the manager. This command may only be executed when the network is not running. If the trigger flag is false, the manager sets the specified time as soon as it receives the setTime command. When the manager receives a Time Pin trigger, it temporarily stores the current time. If a setTime request is received within a short period of time following the trigger, the manager calculates the delay since the trigger and adjust the time such that the trigger was received at the specified time value.',
            'request'    : [
                ['trigger',                 INT,      1,   None],
                ['utcSecs',                 INT,      8,   None],
                ['utcUsecs',                INT,      4,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed. The manager is ready to set the time.',
               'RC_INVALID_ARGUMENT'   : 'One of the parameters was invalid',
               'RC_VALIDATION_ERROR'   : 'Network is running, setTime command is disabled.',
            },
        },
        {
            'id'         : 55,
            'name'       : 'getLicense',
            'description': 'The getLicense command has been deprecated in Manager >= 1.3.0.There is no need to use a license to enable > 32 mote networks.\n\nThe getLicense command returns the current license key.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['license',             HEXDATA,  13,  None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 56,
            'name'       : 'setLicense',
            'description': 'The setLicense command has been deprecated in Manager >= 1.3.0. There is no longer a need to use a license to enable > 32 mote networks.\n\nThe setLicense command validates and updates the software license key stored in flash. Features enabled or disabled by the license key change will take effect after the device is restarted. If the license parameter is set to all 0x0s, the manager restores the default license. This change is persistent.',
            'request'    : [
                ['license',                 HEXDATA,  13,  None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_VALIDATION_ERROR'   : 'The license key is not valid',
               'RC_WRITE_FAIL'         : 'Flash write error, cannot save new settings',
            },
        },
        {
            'id'         : 58,
            'name'       : 'setCLIUser',
            'description': 'The setCLIUser command sets the password that must be used to log into the command line for a particular user role. The user roles are:\n\n- Viewer - read-only access to non-sensitive information\n- User - read-write access This change is persistent.',
            'request'    : [
                ['role',                    INT,      1,   'userRole'],
                ['password',                HEXDATA,  16,  None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_WRITE_FAIL'         : 'Flash write error, can\'t save new settings',
            },
        },
        {
            'id'         : 59,
            'name'       : 'sendIP',
            'description': 'The sendIP command sends a 6LoWPAN packet to a mote in the network. The response contains a callback Id. When the manager injects the packet into the network, it will generate a packetSent notification with the calllbackId. The application is responsible for constructing a valid 6LoWPAN packet. The packet is sent to the mote best-effort, so the application should deal with responses and timeouts, if any.\n\nThe sendIP command should be used by applications that require end-to-end IP connectivity. For applications that do not require end-to-end IP connectivity, the sendData command provides a simpler interface without requiring the application to understand 6LoWPAN encapsulation. For a more comprehensive discussion of the distinction, see the SmartMesh IP Network User Guide.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['priority',                INT,      1,   'packetPriority'],
                ['options',                 INT,      1,   None],
                ['encryptedOffset',         INT,      1,   None],
                ['data',                    HEXDATA,  None,None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['callbackId',          INT,      4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Specified mote is not found',
               'RC_INV_STATE'          : 'Mote is not in operational state',
               'RC_NACK'               : 'User commands queue is full or could not allocate memory buffer for payload',
               'RC_INVALID_ARGUMENT'   : 'Payload size exceeds maximum allowed value or the 6LoWPAN packet is invalid',
            },
        },
        {
            'id'         : 61,
            'name'       : 'restoreFactoryDefaults',
            'description': 'The restoreFactoryDefaults command restores the default configuration and clears the ACL. This change is persistent.\n\nFor Manager versions <1.3.0 that required a license, the license used to enable optional features is preserved during a restore.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_WRITE_FAIL'         : 'Flash write error; cannot save new settings',
            },
        },
        {
            'id'         : 62,
            'name'       : 'getMoteInfo',
            'description': 'The getMoteInfo command returns dynamic information for the specified mote.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['macAddress',          HEXDATA,  8,   None],
                    ['state',               INT,      1,   'moteState'],
                    ['numNbrs',             INT,      1,   None],
                    ['numGoodNbrs',         INT,      1,   None],
                    ['requestedBw',         INT,      4,   None],
                    ['totalNeededBw',       INT,      4,   None],
                    ['assignedBw',          INT,      4,   None],
                    ['packetsReceived',     INT,      4,   None],
                    ['packetsLost',         INT,      4,   None],
                    ['avgLatency',          INT,      4,   None],
                    ['stateTime',           INT,      4,   None],
                    ['numJoins',            INT,      1,   None],
                    ['hopDepth',            INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'No such mote',
            },
        },
        {
            'id'         : 63,
            'name'       : 'getNetworkConfig',
            'description': 'The getNetworkConfig command returns general network configuration parameters, including the Network ID, bandwidth parameters and number of motes.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['networkId',           INT,      2,   None],
                    ['apTxPower',           INTS,     1,   None],
                    ['frameProfile',        INT,      1,   True],
                    ['maxMotes',            INT,      2,   None],
                    ['baseBandwidth',       INT,      2,   None],
                    ['downFrameMultVal',    INT,      1,   None],
                    ['numParents',          INT,      1,   None],
                    ['ccaMode',             INT,      1,   True],
                    ['channelList',         INT,      2,   None],
                    ['autoStartNetwork',    BOOL,     1,   None],
                    ['locMode',             INT,      1,   None],
                    ['bbMode',              INT,      1,   'backboneFrameMode'],
                    ['bbSize',              INT,      1,   None],
                    ['isRadioTest',         INT,      1,   None],
                    ['bwMult',              INT,      2,   None],
                    ['oneChannel',          INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 64,
            'name'       : 'getNetworkInfo',
            'description': 'The getNetworkInfo command returns dynamic network information and statistics.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['numMotes',            INT,      2,   None],
                    ['asnSize',             INT,      2,   None],
                    ['advertisementState',  INT,      1,   True],
                    ['downFrameState',      INT,      1,   'downstreamFrameMode'],
                    ['netReliability',      INT,      1,   None],
                    ['netPathStability',    INT,      1,   None],
                    ['netLatency',          INT,      4,   None],
                    ['netState',            INT,      1,   'networkState'],
                    ['ipv6Address',         HEXDATA,  16,  None],
                    ['numLostPackets',      INT,      4,   None],
                    ['numArrivedPackets',   INT,      8,   None],
                    ['maxNumbHops',         INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 65,
            'name'       : 'getMoteConfigById',
            'description': 'The getMoteConfigById command returns a single mote description as the response. The command takes one argument, the short address of a mote (Mote ID). The command returns the same response structure as the getMoteConfig command.',
            'request'    : [
                    ['moteId',              INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['macAddress',          HEXDATA,  8,   None],
                    ['moteId',              INT,      2,   None],
                    ['isAP',                BOOL,     1,   None],
                    ['state',               INT,      1,   'moteState'],
                    ['reserved',            INT,      1,   None],
                    ['isRouting',           BOOL,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'No such mote',
            },
        },
        {
            'id'         : 66,
            'name'       : 'setCommonJoinKey',
            'description': 'The setCommonJoinKey command will set a new value for the common join key. The common join key is used to decrypt join messages only if the ACL is empty.',
            'request'    : [
                ['key',                     HEXDATA,  16,  None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 67,
            'name'       : 'getIPConfig',
            'description': 'The getIPConfig command returns the manager\'s IP configuration parameters, including the IPv6 address and mask.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['ipv6Address',         HEXDATA,  16,  None],
                    ['mask',                HEXDATA,  16,  None],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
            },
        },
        {
            'id'         : 68,
            'name'       : 'setIPConfig',
            'description': 'The setIPConfig command sets the IPv6 prefix of the mesh network. Only the upper 8 bytes of the IPv6 address are relevant: the lower 8 bytes of the IPv6 address are ignored, and lower 8 bytes of the mask field are reserved and should be set to 0. This change is persistent.',
            'request'    : [
                ['ipv6Address',             HEXDATA,  16,  None],
                ['mask',                    HEXDATA,  16,  None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_WRITE_FAIL'         : 'Flash write error, can\'t save new settings',
            },
        },
        {
            'id'         : 69,
            'name'       : 'deleteMote',
            'description': 'The deleteMote command deletes a mote from the manager\'s list. A mote can only be deleted if it is in the Lost state.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                 : 'Command successfully completed',
               'RC_NOT_FOUND'          : 'Specified mote is not found',
               'RC_INV_STATE'          : 'Mote state is not Lost or mote is access point',
               'RC_WRITE_FAIL'         : 'Flash write error, can\'t save new settings',
            },
        },
        {
            'id'         : 70,
            'name'       : 'getMoteLinks',
            'description': 'The getMoteLinks command returns information about links assigned to the mote. The response contains a list of links starting with Nth link on the mote, where N is supplied as the idx parameter in the request. To retrieve all links on the device the user can call this command with idx that increments by number of links returned with prior response, until the command returns RC_END_OF_LIST response code. Note that links assigned to a mote may change between API calls.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
                ['idx',                     INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['idx',                 INT,      2,   None],
                    ['utilization',         INT,      1,   None],
                    ['numLinks',            INT,      1,   None],
                    ['links',               ARRAY,    10,
                        [
                            ['frameId',         INT,      1,   None],
                            ['slot',            INT,      4,   None],
                            ['channelOffset',   INT,      1,   None],
                            ['moteId',          INT,      2,   None],
                            ['flags',           INT,      1,   None],
                        ]
                    ],
                ],
            },
            'responseCodes': {
               'RC_NOT_FOUND'          : 'No such mote.',
               'RC_INV_STATE'          : 'Mote is not in operational state',
               'RC_END_OF_LIST'        : 'The index requested is greater than number of links.',
            },
        },
        {
            'id'         : 74,
            'name'       : 'radiotestRxPER',
            'description': 'The radiotestRxPER command initiates the Packet Error Rate (PER) test in RX mode. This command may be issued only if the manager has been booted up in radiotest mode (see setNetworkConfig command).\n\nThis command is available in the SmartMesh IP Manager version 1.4.2 or later.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'             : 'Command was accepted',
               'RC_IN_PROGRESS'  : 'The mote is in invalid state to start PER test',
            },
        },
        {
            'id'         : 75,
            'name'       : 'radiotestTxPER',
            'description': 'The radiotestTxPER command initiates the Packet Error Rate (PER) test in TX mode. This command may be issued only if the manager has been booted up in radiotest mode (see setNetworkConfig command).\nChannel numbering is 0-15, corresponding to IEEE 2.4 GHz channels 11-26.\n\nThis command is available in the SmartMesh IP Manager version 1.4.2 or later.',
            'request'    : [
                ['txPower',        INTS,      1,   None],
                ['numPackets',     INT,       2,   None],
                ['chanMask',       HEXDATA,   2,   None],
                ['numRepeat',     INT,       2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'             : 'Command was accepted',
               'RC_IN_PROGRESS'  : 'The mote is in invalid state to start PER test',
            },
        },
    ]
    
    subCommandsEvents = [
        {
            'id'         : 0,
            'name'       : 'eventMoteReset',
            'description': 'This notification is sent when a user-initiated reset is executed by the manager.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',         HEXDATA,  8,   None],
                ],
            },
        },
        {
            'id'         : 1,
            'name'       : 'eventNetworkReset',
            'description': 'This notification is sent when the manager starts the network. This event has no eventData fields.',
            'response'   : {
                'FIELDS':  [
                ],
            },
        },
        {
            'id'         : 2,
            'name'       : 'eventCommandFinished',
            'description': 'The commandFinished notification is sent when a command associated with the provided callback id finishes executing.',
            'response'   : {
                'FIELDS':  [
                    ['callbackId',          HEXDATA,  4,   None],
                    ['rc',                  INT,      1,   'commandFinishedResult'],
                ],
            },
        },
        {
            'id'         : 3,
            'name'       : 'eventMoteJoin',
            'description': 'This notification is sent when a mote joins the network.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',         HEXDATA,  8,   None],
                ],
            },
        },
        {
            'id'         : 4,
            'name'       : 'eventMoteOperational',
            'description': 'This notification is sent when a mote that joins the network becomes operational.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',         HEXDATA,  8,   None],
                ],
            },
        },
        {
            'id'         : 5,
            'name'       : 'eventMoteLost',
            'description': "This notification is sent when a mote's state changes to Lost , which indicates that the mote is not responding to downstream messages.",
            'response'   : {
                'FIELDS':  [
                    ['macAddress',         HEXDATA,  8,   None],
                ],
            },
        },
        {
            'id'         : 6,
            'name'       : 'eventNetworkTime',
            'description': 'The time notification is triggered by the client asserting the TIME pin or by calling the getTime command. This notification contains the time when the TIME pin was asserted (or the getTime command was processed) expressed as:\n\n- ASN The absolute slot number (the number of timeslots since " 7/2/2002 8:00:00 PM PST" if UTC is set on manager, otherwise since Jan 1, 1970)\n\n\n- Uptime The number of seconds since the device was booted\n- Unix time The number of seconds and microseconds since Jan 1, 1970 in UTC',
            'response'   : {
                'FIELDS':  [
                    ['uptime',             INT,      4,   None],
                    ['utcSecs',            INT,      8,   None],
                    ['utcUsecs',           INT,      4,   None],
                    ['asn',                HEXDATA,  5,   None],
                    ['asnOffset',          INT,      2,   None],
                ],
            },
        },
        {
            'id'         : 7,
            'name'       : 'eventPingResponse',
            'description': 'This notification is sent when a reply is received from a mote ping.',
            'response'   : {
                'FIELDS':  [
                    ['callbackId',         INT,      4,   None],
                    ['macAddress',         HEXDATA,  8,   None],
                    ['delay',              INT,      4,   None],
                    ['voltage',            INT,      2,   None],
                    ['temperature',        INTS,     1,   None],
                ],
            },
        },
        {
            'id'         : 10,
            'name'       : 'eventPathCreate',
            'description': 'This notification is sent when the manager creates a connection (path) between two motes.',
            'response'   : {
                'FIELDS':  [
                    ['source',             HEXDATA,  8,   None],
                    ['dest',               HEXDATA,  8,   None],
                    ['direction',          INT,      1,   'pathDirection'],
                ],
            },
        },
        {
            'id'         : 11,
            'name'       : 'eventPathDelete',
            'description': 'This notification is sent when the manager removes a connection (path) between two motes.',
            'response'   : {
                'FIELDS':  [
                    ['source',             HEXDATA,  8,   None],
                    ['dest',               HEXDATA,  8,   None],
                    ['direction',          INT,      1,   'pathDirection'],
                ],
            },
        },
        {
            'id'         : 12,
            'name'       : 'eventPacketSent',
            'description': 'The packetSent notification is generated when client\'s packet is removed from manager\'s queue and sent into the wireless network.',
            'response'   : {
                'FIELDS':  [
                    ['callbackId',         INT,      4,   None],
                    ['rc',                 INT,      1,   None],
                ],
            },
        },
        {
            'id'         : 13,
            'name'       : 'eventMoteCreate',
            'description': 'This event is sent when a mote joins the manager for the first time.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',        HEXDATA,  8,   None],
                    ['moteId',             INT,      2,   None],
                ],
            },
        },
        {
            'id'         : 14,
            'name'       : 'eventMoteDelete',
            'description': 'This notification is sent when a mote is deleted as a result of moteDelete command.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',        HEXDATA,  8,   None],
                    ['moteId',             INT,      2,   None],
                ],
            },
        },
        {
            'id'         : 15,
            'name'       : 'eventJoinFailed',
            'description': 'The joinFailed event is generated when a mote sends a join request to the manager but the request can not be validated. This notification is available in Manager 1.4.1 or later.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',        HEXDATA,  8,   None],
                    ['reason',            INT,      1,   'joinFailedReason'],
                ],
            },
        },
        {
            'id'         : 16,
            'name'       : 'eventInvalidMIC',
            'description': 'The invalidMIC event is generated when a packet that the manager receives from a mote in the network fails decryption. This notification is available in Manager 1.4.1 or later.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',        HEXDATA,  8,   None],
                ],
            },
        },
    ]
    
    subCommandsNotification = [
        {
            'id'         : 1,
            'name'       : 'notifEvent',
            'description': 'Event notification',
            'response'   : {
                'FIELDS':  [
                    ['eventId',            INT,      4,   None],
                    [SUBID2,               INT,      1,   None],
                ],
            },
            'subCommands': subCommandsEvents,
        },
        {
            'id'         : 2,
            'name'       : 'notifLog',
            'description': 'A log notifications is generated in response to the getLog command. Each log notification contains a message from the mote\'s log.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',         HEXDATA,  8,   None],
                    ['logMsg',             HEXDATA,  None,None],
                ],
            },
        },
        {
            'id'         : 4,
            'name'       : 'notifData',
            'description': 'The data notification contains a header and a variable length array of binary data. The length of the data is determined based on the length of the notification.\n\nThe manager forwards all packets received on its IP address and non-manager ports as data notifications.',
            'response'   : {
                'FIELDS':  [
                    ['utcSecs',            INT,      8,   None],
                    ['utcUsecs',           INT,      4,   None],
                    ['macAddress',         HEXDATA,  8,   None],
                    ['srcPort',            INT,      2,   None],
                    ['dstPort',            INT,      2,   None],
                    ['data',               HEXDATA,  None,None],
                ],
            },
        },
        {
            'id'         : 5,
            'name'       : 'notifIpData',
            'description': "The ipData notification contains full IP packet sent by the mote, including 6LoWPAN header, UDP header, and the UDP payload. Manager generates this notification when it receives packet from a mote with destination other than manager's own IP address. The size of the data field can be calculated by subtracting the fixed header size (up to macAddress) from the size of overall notification packet.",
            'response'   : {
                'FIELDS':  [
                    ['utcSecs',            INT,      8,   None],
                    ['utcUsecs',           INT,      4,   None],
                    ['macAddress',         HEXDATA,  8,   None],
                    ['data',               HEXDATA,  None,None],
                ],
            },
        },
        {
            'id'         : 6,
            'name'       : 'notifHealthReport',
            'description': 'The healthReport notifications include the raw payload of health reports received from devices. The payload contains one or more specific health report messages. Each message contains an identifier, length and variable-sized data. The individual healthReport message structures are defined below.',
            'response'   : {
                'FIELDS':  [
                    ['macAddress',         HEXDATA,  8,   None],
                    ['payload',            HEXDATA,  None,None],
                ],
            },
        },
        {
            'id'         : 7,
            'name'       : 'notifRadiotestStatsPER',
            'description': 'The testRadioStatsPER notification is generated by the manager when PER test in RX mode is completed.\n\nThis command is available in the SmartMesh IP Manager version 1.4.2 or later.',
            'response'   : {
                'FIELDS':  [
                    ['numRxOK',           INT,      2,   None],
                    ['numRxErr',          INT,      2,   None],
                    ['numRxInv',          INT,      2,   None],
                    ['numRxMiss',         INT,      2,   None],
                    ['perInt',            INT,      2,   None],
                    ['perFrac',           INT,      2,   None],
                ],
            },
        },
    ]
    
    # We redefine this attribute inherited from ApiDefinition. See
    # ApiDefinition for a full description of the structure of this field.
    notifications = [
        {
            'id'         : 3,
            'name'       : 'manager_hello',
            'description': 'Sent by the manager to a initiate new session with a client.',
            'response'   : {
                'FIELDS':  [
                    ['version',             INT,      1,   None],
                    ['mode',                INT,      1,   None],
                ],
            },
        },
        {
            'id'         : 20,
            'name'       : 'notification',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    [SUBID1,                INT,      1,   None],
                ]
            },
            'subCommands': subCommandsNotification,
        },
    ]
    
    def serialize_radiotestTx(self, commandArray, cmd_params):
        
        # start by serializing with complete API definition
        returnVal = self.default_serializer(commandArray, cmd_params)
        
        # remove the unused sequence definitions
        startIndex = 7+cmd_params['seqSize']*3
        numVals    = 10-cmd_params['seqSize']
        returnVal[1][startIndex:startIndex+numVals*3] = []
        
        return returnVal



#!/usr/bin/python

#============================ define ==========================================

#============================ imports =========================================

import os
import collections
import time
import datetime
import threading
import copy
import pickle
import traceback

if os.name=='nt':       # Windows
   import _winreg as winreg
elif os.name=='posix':  # Linux
   import glob

import json

from SmartMeshSDK                      import sdk_version
from SmartMeshSDK.utils                import FormatUtils as u, \
                                              SerialScanner
from SmartMeshSDK.IpMgrConnectorSerial import IpMgrConnectorSerial
from SmartMeshSDK.IpMgrConnectorMux    import IpMgrSubscribe
from SmartMeshSDK.ApiException         import APIError,      \
                                              ConnectionError
from SmartMeshSDK.protocols.Hr         import HrParser
from SmartMeshSDK.protocols.oap        import OAPDispatcher, \
                                              OAPClient,     \
                                              OAPMessage,    \
                                              OAPDefines as oapdefs

#============================ helpers =========================================

def currentUtcTime():
    return time.strftime("%a, %d %b %Y %H:%M:%S UTC", time.gmtime())

def logCrash(threadName,err):
    output  = []
    output += ["============================================================="]
    output += [currentUtcTime()]
    output += [""]
    output += ["CRASH in Thread {0}!".format(threadName)]
    output += [""]
    output += ["=== exception type ==="]
    output += [str(type(err))]
    output += [""]
    output += ["=== traceback ==="]
    output += [traceback.format_exc()]
    output  = '\n'.join(output)
    
    print output

def reversedict(d):
    return dict((v,k) for (k,v) in d.iteritems())

def stringifyMacIpAddresses(indict):
    '''
    in: {
        'field1':     123,
        'macAddress': [0,1,2,3,4,5,6,7],
    }
    out: {
        'field1':     123,
        'macAddress': '00-01-02-03-04-05-06-07',
    }
    '''
    outdict = indict
    for name in ['macAddress','source','dest']:
        try:
            assert len(outdict[name])==8
            outdict[name] = u.formatMacString(outdict[name])
        except KeyError:
            pass
    for name in ['ipv6Address']:
        try:
            assert len(outdict[name])==16
            outdict[name] = u.formatIpString(outdict[name])
        except KeyError:
            pass
    return outdict

def destringifyMacAddresses(d):
    '''
    in: {
        'field1':     123,
        'macAddress': '00-01-02-03-04-05-06-07',
    }
    out: {
        'field1':     123,
        'macAddress': [0,1,2,3,4,5,6,7],
    }
    returns whether was destringified
    '''
    wasDestringified = False
    for name in ['macAddress','source','dest']:
        try:
            if type(d[name]) in [str,unicode]:
                d[name] = [int(b,16) for b in d[name].split('-')]
                wasDestringified = True
        except KeyError:
            pass
    return wasDestringified

#============================ classes =========================================

class ManagerHandler(threading.Thread):
    '''
    \brief Connects to the manager, re-connects automatically
    '''
    
    def __init__(self,serialport,notifHandler):

        # store params
        self.serialport      = serialport
        self.notifHandler    = notifHandler
        
        # local variables
        self.reconnectEvent  = threading.Event()
        self.dataLock        = threading.RLock()
        self.connector       = None
        self.goOn            = True

        # start the thread
        threading.Thread.__init__(self)
        self.name            = 'ManagerHandler@{0}'.format(self.serialport)
        self.start()
    
    def run(self):
        try:
            while self.goOn:
                try:
                    
                    # connect to the manager
                    self.connector = IpMgrConnectorSerial.IpMgrConnectorSerial()
                    self.connector.connect({
                        'port': self.serialport,
                    })

                    # subscribe to notifications
                    self.subscriber = IpMgrSubscribe.IpMgrSubscribe(self.connector)
                    self.subscriber.start()
                    self.subscriber.subscribe(
                        notifTypes =    [
                                            IpMgrSubscribe.IpMgrSubscribe.NOTIFDATA,
                                        ],
                        fun =           self._notifAll,
                        isRlbl =        False,
                    )
                    self.subscriber.subscribe(
                        notifTypes =    [
                                            IpMgrSubscribe.IpMgrSubscribe.NOTIFEVENT,
                                            IpMgrSubscribe.IpMgrSubscribe.NOTIFHEALTHREPORT,
                                            IpMgrSubscribe.IpMgrSubscribe.NOTIFIPDATA,
                                            IpMgrSubscribe.IpMgrSubscribe.NOTIFLOG,
                                        ],
                        fun =           self._notifAll,
                        isRlbl =        True,
                    )
                    self.subscriber.subscribe(
                        notifTypes =    [
                                            IpMgrSubscribe.IpMgrSubscribe.ERROR,
                                            IpMgrSubscribe.IpMgrSubscribe.FINISH,
                                        ],
                        fun =           self._notifErrorFinish,
                        isRlbl =        True,
                    )

                except Exception as err:
                    
                    try:
                        self.connector.disconnect()
                    except Exception:
                        pass

                    # wait to reconnect
                    time.sleep(1)

                else:
                    self.reconnectEvent.clear()
                    self.reconnectEvent.wait()
                    
                    try:
                        self.connector.disconnect()
                    except Exception:
                        pass

        except Exception as err:
            logCrash(self.name,err)
    
    #======================== public ==========================================
    
    def close(self):

        try:
            self.connector.disconnect()
        except Exception:
            pass

        self.goOn = False
    
    def isConnected(self):
        try:
            return self.connector.isConnected
        except AttributeError:
            return False
    
    #======================== private =========================================

    #=== Dust API notifications
    
    def _notifAll(self, notif_name, dust_notif):
        
        try:
            self.notifHandler(
                self.serialport,
                notif_name,
                dust_notif,
            )
        except Exception as err:
            logCrash(self.name,err)

    def _notifErrorFinish(self,notifName,dust_notif):

        try:
            assert notifName in [
                IpMgrSubscribe.IpMgrSubscribe.ERROR,
                IpMgrSubscribe.IpMgrSubscribe.FINISH,
            ]

            if not self.reconnectEvent.isSet():
                self.reconnectEvent.set()
        except Exception as err:
            logCrash(self.name,err)

class SnapshotThread(threading.Thread):
    '''
    \brief one instance per JsonManager, waits to be triggered, does snapshot on one manager
    '''
    
    def __init__(self,raw_POST,notifCb):
        
        # store params
        self.raw_POST             = raw_POST
        self.notifCb              = notifCb
        
        # local variable
        self.dataLock             = threading.RLock()
        self.snapshotsTodo        = []
        self.snapshotnowSem       = threading.Semaphore(0)
        self.lastsnapshots        = {}
        
        # start the thread
        threading.Thread.__init__(self)
        self.name                 = 'SnapshotThread'
        self.daemon               = True
        self.start()
    
    def run(self):
        try:
            while True:
                
                # wait for trigger
                self.snapshotnowSem.acquire()
                
                # get the manager and correlationID
                with self.dataLock:
                    (manager,correlationID) = self.snapshotsTodo.pop(0)
                
                # do the snapshot
                try:
                    snapshot = {}
                    
                    # timestamp_start
                    snapshot['timestamp_start'] = currentUtcTime()
                    
                    # getSystemInfo()
                    resp = self.raw_POST(
                        commandArray   = ["getSystemInfo"],
                        fields         = {},
                        manager        = manager,
                    )
                    snapshot['getSystemInfo'] = stringifyMacIpAddresses(resp)
                    
                    # getNetworkConfig()
                    resp = self.raw_POST(
                        commandArray   = ["getNetworkConfig"],
                        fields         = {},
                        manager        = manager,
                    )
                    snapshot['getNetworkConfig'] = stringifyMacIpAddresses(resp)
                    
                    # getNetworkInfo()
                    resp = self.raw_POST(
                        commandArray   = ["getNetworkInfo"],
                        fields         = {},
                        manager        = manager,
                    )
                    snapshot['getNetworkInfo'] = stringifyMacIpAddresses(resp)
                    
                    # getMoteConfig() on all motes
                    snapshot['getMoteConfig'] = {}
                    macs       = []
                    currentMac = [0]*8
                    while True:
                        resp = self.raw_POST(
                            commandArray   = ["getMoteConfig"],
                            fields         = {
                                "macAddress": currentMac,
                                "next": True
                            },
                            manager        = manager,
                        )
                        if resp['RC'] != 0:
                            break
                        mac           = resp['macAddress']
                        macString     = u.formatMacString(mac)
                        snapshot['getMoteConfig'][macString] = stringifyMacIpAddresses(resp)
                        macs         += [mac]
                        currentMac    = mac
                    
                    # getMoteInfo() on all motes
                    snapshot['getMoteInfo'] = {}
                    for mac in macs:
                        resp = self.raw_POST(
                            commandArray   = ["getMoteInfo"],
                            fields         = {
                                "macAddress": mac,
                            },
                            manager        = manager,
                        )
                        macString     = u.formatMacString(mac)
                        snapshot['getMoteInfo'][macString] = stringifyMacIpAddresses(resp)
                    
                    # getPathInfo() on all paths on all motes
                    snapshot['getPathInfo'] = {}
                    for mac in macs:
                        macString     = u.formatMacString(mac)
                        snapshot['getPathInfo'][macString] = {}
                        currentPathId  = 0
                        while True:
                            resp = self.raw_POST(
                                commandArray   = ["getNextPathInfo"],
                                fields         = {
                                    "macAddress": mac,
                                    "filter":     0,
                                    "pathId":     currentPathId
                                },
                                manager        = manager,
                            )
                            if resp["RC"] != 0:
                                break
                            snapshot['getPathInfo'][macString][currentPathId] = stringifyMacIpAddresses(resp)
                            currentPathId  = resp["pathId"]
                    
                    # getMoteLinks() on all paths on all motes
                    snapshot['getMoteLinks'] = {}
                    for mac in macs:
                        macString     = u.formatMacString(mac)
                        snapshot['getMoteLinks'][macString] = {}
                        snapshot['getMoteLinks'][macString]['links'] = []
                        currentidx  = 0
                        while True:
                            resp = self.raw_POST(
                                commandArray   = ["getMoteLinks"],
                                fields         = {
                                    "macAddress": mac,
                                    "idx":        currentidx,
                                },
                                manager        = manager,
                            )
                            if resp["RC"] != 0:
                                break
                            # add all "metadata" fields, i.e. every before the list of links
                            for (k,v) in resp.items():
                                if ("_" not in k) and (k not in ['numLinks','idx']):
                                    snapshot['getMoteLinks'][macString][k] = v
                            # populate list of links
                            for i in range(resp['numLinks']):
                                thisLink = {}
                                suffix = '_{0}'.format(i+1)
                                for (k,v) in resp.items():
                                    if k.endswith(suffix):
                                        name = k[:-len(suffix)]
                                        thisLink[name]   = v
                                snapshot['getMoteLinks'][macString]['links'] += [thisLink]
                            currentidx += resp['numLinks']
                    
                    # timestamp_stop
                    snapshot['timestamp_stop'] = currentUtcTime()
                    
                    # epoch_stop
                    snapshot['epoch_stop']     = time.time()
                    
                    # remember the last snapshot for each manager
                    with self.dataLock:
                        self.lastsnapshots[manager] = snapshot
                    
                except Exception as err:
                    notifJson    = {
                        'valid':    False,
                        'err':      str(err),
                    }
                else:
                    notifJson    = {
                        'valid':    True,
                        'snapshot': snapshot,
                    }
                finally:
                    notifJson['manager']              = manager
                    if correlationID:
                        notifJson['correlationID']    = correlationID
                    self.notifCb(
                        notifName    = 'snapshot',
                        notifJson    = notifJson,
                    )
        
        except Exception as err:
            logCrash(self.name,err)

    #======================== public ==========================================
    
    def doSnapshot(self,manager,correlationID):
        with self.dataLock:
            self.snapshotsTodo += [(manager,correlationID)]
        self.snapshotnowSem.release()
    
    def getLastsnapshots(self):
        with self.dataLock:
            returnVal = copy.deepcopy(self.lastsnapshots)
        now = time.time()
        for m in returnVal.keys():
            returnVal[m]['age_seconds'] = int(now-returnVal[m]['epoch_stop'])
        return returnVal

class DeleMgrThread(threading.Thread):
    
    HOUSEKEEPING_PERIOD = 10
    
    def __init__(self,jsonManager):
        # store params
        self.jsonManager = jsonManager
        
        # start thread
        threading.Thread.__init__(self)
        self.name                 = 'DeleMgrThread'
        self.daemon               = True
        self.start()
    
    def run(self):
        while True:
            for (m,c) in self.jsonManager.status_GET()['managers'].items():
                if c=='disconnected':
                    self.jsonManager.managers_DELETE([m])
            time.sleep(self.HOUSEKEEPING_PERIOD)

class JsonManager(object):
    
    OAP_TIMEOUT = 30.000
    
    def __init__(self, autoaddmgr, autodeletemgr, serialport, notifCb, configfilename=None):
        
        # store params
        self.autoaddmgr           = autoaddmgr
        self.autodeletemgr        = autodeletemgr
        self.serialport           = serialport
        self.notifCb              = notifCb
        self.configfilename       = configfilename
        
        # local variables
        self.startTime            = time.time()
        self.dataLock             = threading.RLock()
        self._loadConfig()   # populates self.config dictionnary
        self.managerHandlers      = {}
        self.oapDispatch          = OAPDispatcher.OAPDispatcher()
        self.oapDispatch.register_notif_handler(self._manager_oap_notif_handler)
        self.oapClients           = {}
        self.snapshotThread       = SnapshotThread(
            self.raw_POST,
            self.notifCb,
        )
        self.outstandingEvents    = {}
        self.responses            = {}
        self.hrParser             = HrParser.HrParser()
        
        # connect to managers (if any)
        self._syncManagers()
        
        # if autoaddmgr, have SerialScanner looks for manager
        if self.autoaddmgr:
            self.serialScanner = SerialScanner.SerialScanner()
            self.serialScanner.availableManagerNotifier(
               cb     = self._availablemanagers_cb,
               period = 60,
            )
        
        # if autodeletemgr, start DeleMgrThread
        if self.autodeletemgr:
            DeleMgrThread(self)
    
    #======================== public ==========================================
    
    #=== status
    
    def status_GET(self):
        return {
            'SmartMesh SDK version': '.'.join([str(b) for b in sdk_version.VERSION]),
            'current time':          self._formatTime(),
            'running since':         '{0} ({1} ago)'.format(
                self._formatTime(self.startTime),
                datetime.timedelta(seconds=time.time()-self.startTime)
            ),
            'threads running':       [t.getName() for t in threading.enumerate()],
            'managers':              self._formatManagersStatus(),
        }
    
    #=== raw
    
    def raw_POST(self, commandArray, fields, manager):
        if type(manager)==int:
            manager = sorted(self.managerHandlers.keys())[manager]
        
        # mac addresses: '00-01-02-03-04-05-06-07' -> [0,1,2,3,4,5,6,7]
        wasDestringified = destringifyMacAddresses(fields)
        
        with self.dataLock:
            try:
                returnVal = self.managerHandlers[manager].connector.send(
                    commandArray = commandArray,
                    fields       = fields,
                )
            except APIError as err:
                returnVal = {
                    'RC': err.rc,
                }
        
        if wasDestringified:
            # mac addresses: [0,1,2,3,4,5,6,7] -> '00-01-02-03-04-05-06-07'
            stringifyMacIpAddresses(returnVal)
        
        return returnVal
    
    #=== oap
    
    # /info
    
    def oap_info_GET(self,mac):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.GET,
            resource    = 'info',
        )
    
    # /main
    
    def oap_main_GET(self,mac):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.GET,
            resource    = 'main',
        )
    
    def oap_main_PUT(self,mac,body):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.PUT,
            resource    = 'main',
            body        = body,
        )
    
    # /digital_in
    
    def oap_digital_in_GET(self,mac,pin):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.GET,
            resource    = 'digital_in',
            subresource = pin,
        )
    
    def oap_digital_in_PUT(self,mac,pin,body):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.PUT,
            resource    = 'digital_in',
            subresource = pin,
            body        = body,
        )
    
    # /digital_out
    
    def oap_digital_out_PUT(self,mac,pin,body):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.PUT,
            resource    = 'digital_out',
            subresource = pin,
            body        = body,
        )
    
    # /analog
    
    def oap_analog_GET(self,mac,pin):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.GET,
            resource    = 'analog',
            subresource = pin,
        )
    
    def oap_analog_PUT(self,mac,pin,body):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.PUT,
            resource    = 'analog',
            subresource = pin,
            body        = body,
        )
    
    # /temperature
    
    def oap_temperature_GET(self,mac):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.GET,
            resource    = 'temperature',
        )
    
    def oap_temperature_PUT(self,mac,body):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.PUT,
            resource    = 'temperature',
            body        = body,
        )
    
    # /pkgen
    
    def oap_pkgen_echo_GET(self,mac):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.GET,
            resource    = 'pkgen',
            subresource = 0,
        )
    
    def oap_pkgen_PUT(self,mac,body):
        return self._oap_send_and_wait_for_reply(
            mac         = mac,
            method      = OAPMessage.CmdType.PUT,
            resource    = 'pkgen',
            body        = body,
        )
    
    #=== helpers
    
    def serialports_GET(self):
        try:
            serialports = []
            
            if os.name=='nt':
                key  = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, 'HARDWARE\\DEVICEMAP\\SERIALCOMM')
                for i in range(winreg.QueryInfoKey(key)[1]):
                    try:
                        val = winreg.EnumValue(key,i)
                    except:
                        pass
                    else:
                        if val[0].find('VCP')>-1:
                            serialports.append(str(val[1]))
            elif os.name=='posix':
                serialports = glob.glob('/dev/ttyUSB*')
            
            serialports.sort()
            
            return {
                'serialports': serialports
            }
        except Exception as err:
            return ['Could not scan for serial port. Error={0}'.format(err)]
    
    def motes_GET(self):
        with self.dataLock:
            returnVal = {m:self._list_motes_per_manager(m) for m in self.managerHandlers.keys()}
        return returnVal
    
    def oapmotes_GET(self):
        with self.dataLock:
            return {'oapmotes': self.oapClients.keys(),}
    
    def snapshot_POST(self,manager,correlationID=None):
        self.snapshotThread.doSnapshot(manager,correlationID)
    
    def snapshot_GET(self):
        return self.snapshotThread.getLastsnapshots()
    
    def managers_PUT(self,newmanagers):
        with self.dataLock:
            for m in newmanagers:
                if m not in self.config['managers']:
                    self.config['managers'] += [m]
        self._saveConfig()
        self._syncManagers()
    
    def managers_DELETE(self,oldmanagers):
        with self.dataLock:
            for m in oldmanagers:
                try:
                    self.config['managers'].remove(m)
                except ValueError:
                    pass # happens when manager doesn't exist
        self._saveConfig()
        self._syncManagers()
    
    #=== config
    
    def config_GET(self):
        with self.dataLock:
           return copy.deepcopy(self.config)
    
    def config_POST(self,newconfig):
        with self.dataLock:
            self.config = self._recursive_dict_update(
                self.config,
                newconfig
            )
        self._saveConfig()
        self._syncManagers()
        with self.dataLock:
           return copy.deepcopy(self.config)
    
    #=== close
    
    def close(self):
        for (k,v) in self.managerHandlers.items():
            try:
                v.close()
            except:
                pass
    
    #======================== private =========================================
    
    #=== api
    
    # status
    
    def _formatManagersStatus(self):
        returnVal = {}
        
        with self.dataLock:
            for (k,v) in self.managerHandlers.items():
                if v.isConnected():
                    returnVal[k] = 'connected'
                else:
                    returnVal[k] = 'disconnected'
        
        return returnVal
    
    # oap
    
    def _oap_send_and_wait_for_reply(self, mac, method, resource, subresource=None, body={}):
        
        # use only lowercase
        mac = mac.lower()
        
        # add an oapClient, if needed
        self._oap_add_client_if_needed(mac)
        
        # create data_tags
        data_tags = []
        if method==OAPMessage.CmdType.PUT:
            for (i,(n,t,d)) in enumerate(oapdefs.FIELDS[resource]):
                if d.count('w'):
                    if   t=='INT8U':
                        data_tags += [OAPMessage.TLVByte(   t=i,v=body[n])]
                    elif t=='INT16U':
                        data_tags += [OAPMessage.TLVShort(  t=i,v=body[n])]
                    elif t=='INT32U':
                        data_tags += [OAPMessage.TLVLong(   t=i,v=body[n])]
                    elif t=='INT8U[16]':
                        temp = body[n]
                        temp = ''.join([chr(int(b,16)) for b in [temp[2*j:2*j+2] for j in range(len(temp)/2)]])
                        data_tags += [OAPMessage.TLVString( t=i,v=temp)]
        
        # send OAP request
        addr           = [b for b in oapdefs.ADDRESS[resource]]
        if subresource!=None:
            addr      += [subresource]
        self.oapClients[mac].send(
            cmd_type   = method,
            addr       = addr,
            data_tags  = data_tags,
            cb         = self._oap_handle_response,
        )
        
        # wait for and handle OAP response
        with self.dataLock:
            if mac in self.outstandingEvents:
                raise SystemError('busy waiting for response')
            event = threading.Event()
            self.outstandingEvents[mac] = event
        if event.wait(self.OAP_TIMEOUT):
            # received response
            with self.dataLock:
                response = self.responses[mac]
                del self.responses[mac]
            return self._oap_format_response(resource,response)
        else:
            # timeout
            with self.dataLock:
                del self.outstandingEvents[mac]
            bottle.response.status = 504 # Gateway Timeout
            bottle.response.content_type = 'application/json'
            return json.dumps({
                'body': 'timeout!',
            })
    
    def _oap_handle_response(self, mac, oap_resp):
        macString = u.formatMacString(mac)
        
        with self.dataLock:
            if macString in self.responses:
                raise SystemError('response unread')
            
            if macString in self.outstandingEvents:
                self.responses[macString] = oap_resp
                self.outstandingEvents[macString].set()
                del self.outstandingEvents[macString]
            else:
                # received a response I'm not waiting for anymore (increase timeout?)'
                pass
    
    def _oap_format_response(self,resource,response):
        '''
        {
            'command': 1,
            'result': 0,
            'tags': [
                (255, 1, array('B', [0])),
                (0, 1, array('B', [1])),
                (1, 1, array('B', [3])),
                (2, 1, array('B', [0])),
                (3, 2, array('B', [0, 24])),
                (4, 2, array('B', [0, 1])),
                (5, 4, array('B', [0, 0, 0, 0])),
                (6, 4, array('B', [0, 0, 0, 0]))
            ]
        }
        {
            'resource': [0],
            'method':   1,
            'status':   0,
            'fields': {
                0: [1],
                1: [3],
                2: [0],
                3: [0, 24],
                4: [0, 1],
                5: [0, 0, 0, 0],
                6: [0, 0, 0, 0]
            }
        }
        '''
        tags = {k:v.tolist() for (k,_,v) in response['tags']}
        returnVal = {}
        # status
        returnVal['status']       = reversedict(oapdefs.RC)[response['result']]
        # resource
        returnVal['resource']     = reversedict(oapdefs.ADDRESS)[tuple(tags[255])]
        del tags[255]
        # method
        returnVal['method']       = reversedict(oapdefs.COMMAND)[response['command']]
        # fields
        desc = oapdefs.FIELDS[resource]
        fields = {}
        for (k,v) in tags.items():
            (n,t,d)               = desc[k]
            if t=='INT8U[16]':
                fields[n]         = ''.join(["%.2x"%i for i in v])
            elif v==[]:
                fields[n]         = v
            else:
                fields[n]         = int(''.join([chr(b) for b in v]).encode('hex'), 16)
        returnVal['fields']       = fields
        return returnVal
    
    def _oap_add_client_if_needed(self,mac):
        if type(mac)==str:
            macString = mac
            mac       = [int(b,16) for b in mac.split('-')]
        else:
            macString = u.formatMacString(mac)
        
        with self.dataLock:
            if macString not in self.oapClients:
                # get MACs per manager
                for (manager,motes) in self.motes_GET().items():
                    if macString in manager:
                       break
                # create OAPClient
                self.oapClients[macString] = OAPClient.OAPClient(
                    mac,
                    self.managerHandlers[manager].connector.dn_sendData,
                    self.oapDispatch,
                )
    
    # helpers
    
    def _list_motes_per_manager(self,manager):
        returnVal = []
        
        try:
            currentMac     = (0,0,0,0,0,0,0,0) # start getMoteConfig() iteration with the 0 MAC address
            continueAsking = True
            while continueAsking:
                try:
                    with self.dataLock:
                        res = self.managerHandlers[manager].connector.dn_getMoteConfig(currentMac,True)
                except APIError:
                    continueAsking = False
                else:
                    if ((not res.isAP) and (res.state in [4,])):
                        returnVal.append(u.formatMacString(res.macAddress))
                    currentMac = res.macAddress
        except ConnectionError as err:
            pass # happens when manager is disconnected
        
        return returnVal
    
    # manager management
    
    def _syncManagers(self):
        with self.dataLock:
            # add
            for m in self.config['managers']:
                if m not in self.managerHandlers:
                    self.managerHandlers[m] = ManagerHandler(m,self._manager_raw_notif_handler)
            # remove
            for m in self.managerHandlers.keys():
                if m not in self.config['managers']:
                    self.managerHandlers[m].close()
                    del self.managerHandlers[m]
    
    def _availablemanagers_cb(self,serialport):
        self.managers_PUT([serialport])
    
    # config
    
    def _loadConfig(self):
        with self.dataLock:
            try:
                if self.configfilename:
                    self.config = pickle.load(open(self.configfilename,"rb"))
                else:
                   raise IOError
            except IOError as err:
                if self.serialport:
                    managers = [self.serialport]
                else:
                    managers = []
                self.config     = {
                    'managers':         managers,
                    'notification_urls': {
                        'event':            [
                            'http://127.0.0.1:1880/event',
                            'http://127.0.0.1:8081/event',
                        ],
                        'notifLog':         [
                            'http://127.0.0.1:1880/notifLog',
                        ],
                        'notifData':        [
                            'http://127.0.0.1:1880/notifData',
                        ],
                        'notifIpData':      [
                            'http://127.0.0.1:1880/notifIpData',
                        ],
                        'notifHealthReport':[
                            'http://127.0.0.1:1880/notifHealthReport',
                        ],
                        'oap':              [
                            'http://127.0.0.1:1880/oap',
                        ],
                        'hr':               [
                            'http://127.0.0.1:1880/hr',
                        ],
                        'snapshot':         [
                            'http://127.0.0.1:1880/snapshot',
                            'http://127.0.0.1:8081/snapshot',
                        ],
                    }
                }
                self._saveConfig()
    
    def _saveConfig(self):
        if self.configfilename:
            with self.dataLock:
                pickle.dump(self.config, open(self.configfilename,"wb"))
    
    #=== notifications
    
    def _manager_raw_notif_handler(self,manager,notifName,notif):
        
        # parse further
        if   notifName==IpMgrSubscribe.IpMgrSubscribe.NOTIFDATA:
            # try to parse data notifications as OAP (fails if not OAP payload, no problem)
            self.oapDispatch.dispatch_pkt(notifName,notif)
        elif notifName==IpMgrSubscribe.IpMgrSubscribe.NOTIFHEALTHREPORT:
            hr  = self.hrParser.parseHr(notif.payload)
            # POST HR to some URL
            self.notifCb(
                notifName    = 'hr',
                notifJson    = {
                    'name':    'hr',
                    'mac':     u.formatMacString(notif.macAddress),
                    'hr':      hr,
                },
            )
        
        # POST raw notification to some URL
        if notifName.startswith('event'):
            nm           = 'event'
        else:
            nm           = notifName
        fields = stringifyMacIpAddresses(notif._asdict())
        self.notifCb(
            notifName    = nm,
            notifJson    = {
                'manager': manager,
                'name':    notifName,
                'fields':  fields,
            },
        )
    
    def _manager_oap_notif_handler(self,mac,notif):
        
        macString = u.formatMacString(mac)
        
        # add an oapClient, if needed
        self._oap_add_client_if_needed(mac)
        
        # POST OAP notification to some URLs
        fields = stringifyMacIpAddresses(notif._asdict())
        self.notifCb(
            notifName    = 'oap',
            notifJson    = {
                'name':    'oap',
                'mac':     macString,
                'fields':  fields,
            },
        )
    
    #=== formatting
    
    def _formatTime(self,ts=None):
        return time.strftime("%m/%d/%Y %H:%M:%S",time.localtime(ts))
    
    #=== helpers
    
    def _recursive_dict_update(self,d,u):
        for k, v in u.items():
            if isinstance(v, collections.Mapping):
                r = self._recursive_dict_update(d.get(k, {}), v)
                d[k] = r
            else:
                d[k] = u[k]
        return d



"""
This module defines the classes that supports the common structure
for ML Tasks of assets.
"""

from pydantic import BaseModel, Field
from typing import List, Dict, Union, Optional
from enum import Enum


class TaskEventTaxonomyEnum(Enum):
    """An event produced by a task, can be classified within this
    taxonomy.

    For example, for a ML task that recognizes the normal operation
    of a machine, the prediction output could be 'NormalOperation'.
    This output could be categorized as `Harmless`.
    """
    Detrimental = 'Detrimental'
    Harmless = 'Harmless'
    Neutral = 'Neutral'


class ClassificationLabelDefinition(BaseModel):
    labels: List[str]
    label_to_int: Dict[str, int]
    int_to_label: Dict[int, str]
    label_to_taxonomy: Dict[str, TaskEventTaxonomyEnum]


class AssetTask(BaseModel):
    """
    Args:
        label_def: The definition for the event labels of the task.

        task_id: A unique task_id under a given tenant asset.

        asset_id: The asset_id to which this task is associated.

        task_data: A string path in storage for the task model data.

        edges: The list of edges logical ids that feed this tasks as
            signal inputs.

        schema_: A Dictionary of Edges to expected Perception name and
            table name.

    A example of a schema would be:

    """
    label_def: ClassificationLabelDefinition
    task_type: str
    task_id: str
    asset_id: str = Field(alias='machine_id')
    task_data: Optional[str]
    edges: List[str]
    schema_: Dict[str, Dict[str, Dict[str, Union[bool, List]]]] = Field(alias='schema')
    parameters: dict

    class Config:
        # https://github.com/samuelcolvin/pydantic/issues/602#issuecomment-503189368
        allow_population_by_field_name = True


class AssetStateCondition(Enum):
    OPTIMAL = "OPTIMAL"
    OPERATIONAL = "OPERATIONAL"
    UNDEFINED = "UNDEFINED"
    CRITICAL = "CRITICAL"
    GOOD = "GOOD"


class AssetState(BaseModel):
    events: List[str]
    asset_id: str
    condition: AssetStateCondition
    observation_timestamp: int
    tenant_id: str
    partition_timestamp: str




"""
This module implements a DriversHost for the ArmRobotDriver.
"""

# ============================ imports =========================================
import multiprocessing
from typing import Tuple, List

from txp.devices.drivers.arm_robot.arm_controller import ArmController
from txp.devices.drivers.arm_robot.arm_handler import ArmHandler
from txp.devices.drivers.arm_robot.thermal_robot import ThermalRobotDriver
from txp.devices.drivers.driver import StateReport, DriverState
from txp.devices.drivers.drivers_host import DriversHost
from txp.common.edge import EdgeDescriptor
from txp.devices.package_queue import PackageQueueProxy

from txp.devices.drivers.arm_robot.robot_driver import ArmRobotDriver
from txp.common.edge import DeviceId
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# RobotDriversHost is the Host for mock drivers used in unit testing.
# ==============================================================================
class RobotDriversHost(DriversHost):
    # Value requested by the super class
    _driver_errored_timeout_seconds = 15

    def __init__(self, gateway_packages_queue: PackageQueueProxy,
                 drivers_state_queue: multiprocessing.Queue, **kwargs):
        """
        Args:
            gateway_packages_queue: The Gateway package queue to be used by Drivers
                to send collected GatewayPackages.
            drivers_state_queue: The queue provided by the Devices Manager to queue
                drivers states reports.
        """
        super(RobotDriversHost, self).__init__(gateway_packages_queue, drivers_state_queue,
                                               self._driver_errored_timeout_seconds)

        self._actuator_edge_decriptor: EdgeDescriptor = None
        self._drivers_edges_descriptor: List[EdgeDescriptor] = []
        self._arm_controller = None

    @staticmethod
    def get_driver_host_name():
        return ArmRobotDriver.device_name()

    def discover_physical_ids(self) -> List[DeviceId]:
        log.debug(f"{self.__class__.__name__} does not detect devices yet.")
        return []

    def connect(self):
        for driver in self.drivers:
            if driver._robotic_arm_controller is None:
                driver._robotic_arm_controller = self._arm_controller
            driver.connect_device()
            log.debug("connecting driver {}".format(driver.logical_id))

    def disconnect(self):
        if self._arm_controller is not None:
            self._arm_controller.disconnect()
            self._arm_controller = None
        for driver in self.drivers:
            driver.disconnect_device()

    def _add_drivers(self, edges: List[EdgeDescriptor]):
        for descriptor in edges:
            if descriptor.is_actuator_device():
                self._actuator_edge_decriptor = descriptor
                if self._arm_controller is None:
                    self._arm_controller = ArmController(descriptor)
                    log.info("{} arm controller instantiated".format(descriptor.logical_id))
            else:
                self._drivers_edges_descriptor.append(descriptor)
                log.info("{} instantiating virtual driver".format(descriptor.logical_id))
                driver = ArmRobotDriver(descriptor, self.gateway_package_queue,self._change_driver_state_callback,
                                        self._arm_controller, self._completed_driver_pulses_queue)
                self.drivers.append(driver)

    def _change_driver_state_callback(self, logical_id: str, driver_state: DriverState):
        driver = list(filter(lambda driver: driver.logical_id == logical_id, self.drivers)).pop()
        driver.update_driver_state(driver_state)
        self._add_drivers_state_to_queue(StateReport(driver.logical_id, driver_state))

    def is_actuator_device(self, logical_id: str) -> bool:
        return logical_id == self._actuator_edge_decriptor.logical_id

    def take_fix_error_action(self, serial):
        pass


class ThermalArmHost(RobotDriversHost):
    def __init__(self, gateway_packages_queue: PackageQueueProxy,
                 drivers_state_queue: multiprocessing.Queue, **kwargs):
        super(ThermalArmHost, self).__init__(gateway_packages_queue, drivers_state_queue)

    @staticmethod
    def get_driver_host_name():
        return ThermalRobotDriver.device_name()

    def _add_drivers(self, edges: List[EdgeDescriptor]):
        for descriptor in edges:
            if descriptor.is_actuator_device():
                self._actuator_edge_decriptor = descriptor
                if self._arm_controller is None:
                    self._arm_controller = ArmController(descriptor)
                    log.info("{} arm controller instantiated".format(descriptor.logical_id))
            else:
                self._drivers_edges_descriptor.append(descriptor)
                log.info("{} instantiating virtual driver".format(descriptor.logical_id))
                driver = ThermalRobotDriver(descriptor, self.gateway_package_queue,self._change_driver_state_callback,
                                        self._arm_controller, self._completed_driver_pulses_queue)
                self.drivers.append(driver)



"""
    This module provides functions and classes that helps with programmatic interaction with
    Identity Platform and related authentication services.
"""
# ============================ imports =========================================
from google_auth_oauthlib import flow
import firebase_admin
from firebase_admin import auth as firebase_auth
from google.oauth2 import service_account
import requests
import json
from typing import Optional, List, Dict, Tuple
import enum
import logging
import dataclasses

# ============================ constants and data ============================
# Info about Google Auth Scopes can be found at:
#   https://developers.google.com/identity/protocols/oauth2/scopes
_DEFAULT_AUTH_SCOPES = [
    "https://www.googleapis.com/auth/cloud-platform",
    "https://www.googleapis.com/auth/userinfo.email",
    "https://www.googleapis.com/auth/userinfo.profile",
    "openid",
]


@dataclasses.dataclass
class AuthenticationInfo:
    """This class holds information of the authentication process supported by the
    provided functions in this module.

    Args:
        identity_platform_credentials: The IdentityPlatform response for the method
            accounts.signInWithIdp. More info at:
                https://cloud.google.com/identity-platform/docs/reference/rest/v1/accounts/signInWithIdp#response-body
        role_service_account_credentials: The Service Account credentials created for the
            user appropriate Role.
    """

    class UserRole(enum.Enum):
        Admin = 'Admin'
        Client = 'Client'

    identity_platform_credentials: Dict = dataclasses.field(default_factory=dict)
    role_service_account_credentials: service_account.Credentials = None
    user_role: UserRole = None

    @property
    def user_email(self) -> Optional[str]:
        """Returns the user email found in the Identity Platform authenticated
        credentials.
        """
        if "email" in self.identity_platform_credentials:
            return self.identity_platform_credentials["email"]
        else:
            logging.warning("No email field found in identity platform credentials.")
            return None

    @property
    def id_token(self) -> Optional[str]:
        """Returns the Identity Platform Token found in the Identity Platform authenticated"""
        if "idToken" in self.identity_platform_credentials:
            return self.identity_platform_credentials["idToken"]
        else:
            logging.warning("No idToken field found in identity platform credentials.")
            return None

    def is_user_logged_in(self) -> bool:
        # Todo: add a more detailed verification
        return bool(self.identity_platform_credentials) and bool(
            self.role_service_account_credentials
        )


# ========================== Functions =======================================
def get_google_oauth_flow(
    client_oauth_secret: str,
    redirect_url: str = "localhost:8501",
    scopes: List[str] = _DEFAULT_AUTH_SCOPES,
) -> flow.Flow:
    """Returns a Google Authentication Flow.

    More info can be found in the API docs:
        https://google-auth-oauthlib.readthedocs.io/en/latest/reference/google_auth_oauthlib.flow.html

    Args:
        client_oauth_secret(str): The OAuth 2.0 client ID for web application.
        redirect_url(str): The redirect URL to be used with authentication flow
        scopes(List[str]): The list of scopes to use in with OAuth2.0 flow.


    """
    try:
        _app_flow = flow.Flow.from_client_config(
            json.loads(client_oauth_secret),
            scopes=scopes,
            redirect_uri=redirect_url,
        )
    except ValueError:
        raise RuntimeError(
            "Client configuration is not in the correct format. Impossible"
            "to create OAuth flow."
        )
    else:
        return _app_flow


def sing_in_user_in_identity_platform(
    auth_code: str,
    firebase_api_key,
    client_oauth_secret: str,
    redirect_url: str = "localhost:8501",
    scopes: List[str] = _DEFAULT_AUTH_SCOPES,
) -> Optional[Dict]:
    """Sign in an user in Identity Platform using the received OAuth2.0 code
    for the user after successful OAuth2.0 authentication.

    More info at:
        https://cloud.google.com/identity-platform/docs/reference/rest/v1/accounts/signInWithIdp

    Args:
        auth_code: The authorization code response returned by Google afer the user
            successfully authenticated himself using the OAuth2.0 Flow
        client_oauth_secret(str): The OAuth 2.0 client ID for web application.
        redirect_url(str): The redirect URL to be used with authentication flow
        scopes(List[str]): The list of scopes to use in with OAuth2.0 flow.
    """
    google_auth_flow = get_google_oauth_flow(client_oauth_secret, redirect_url, scopes)
    google_auth_flow.fetch_token(code=auth_code)
    oauth_credentials = google_auth_flow.credentials

    logging.debug("Access token obtained from Google Auth flow consent")
    logging.debug("Trying to sing in into Identity Platform")

    # Sign in into identity platform
    query_params = {"key": firebase_api_key}
    headers = {"Content-Type": "application/json"}
    data = {
        "postBody": f"access_token={oauth_credentials.token}&providerId=google.com",
        "requestUri": "http://localhost",  # This can be localhost because manual credentials were provided
        "returnIdpCredential": True,
        "returnSecureToken": True,
    }

    r = requests.post(
        "https://identitytoolkit.googleapis.com/v1/accounts:signInWithIdp",
        params=query_params,
        headers=headers,
        data=json.dumps(data),
    )

    if r.status_code == 200:
        identity_platform_credentials = json.loads(r.content.decode())
        logging.info(
            f"User successfully authenticated with Google "
            f'Identity Platform {identity_platform_credentials["fullName"]}'
        )
        return identity_platform_credentials

    else:
        logging.error(
            f"Unexpected Error when authenticating the user "
            f"with Identity Platform: {r.content.decode()}"
        )
        return None


def _get_firebase_default_app(
    firebase_default_service_account: str,
) -> Optional[firebase_admin.App]:
    """Returns the default firebase APP given the Firebase default service account

    If there's some problem with the creation of the App instance, then the streamlit
    app will set the unexpected error flag for the session.

    Args:
        firebase_default_service_account: The default service account from
            Firebase as a string
    """
    try:
        json_acc_info = json.loads(firebase_default_service_account, strict=False)

    except:
        logging.error(
            "There was an error parsing the Firebase default "
            "service account information"
        )
        return None

    try:
        credentials = firebase_admin.credentials.Certificate(json_acc_info)
        firebase_app = firebase_admin.initialize_app(credentials)
    except:
        logging.error(
            "There was an error creating the Firebase Default " "Application instance"
        )
        return None
    else:
        return firebase_app


def _get_user_role(user_email: str) -> Optional[str]:
    """Returns the User Role for the logged user in the state.

    If the user doesn't have a role (claim) assigned in Identity Platform, then
    the app will set the unexpected error flag.

    TODO: Validate the role against a list of expected roles. Add this when the roles functionalities
        and the application use cases for each type of role are being implemented.

    Args:
        user_email(str): The user email address to request the claims for in Identity Platform.
    """
    user = firebase_auth.get_user_by_email(user_email)
    claims = user.custom_claims
    if "role" not in claims:
        logging.error("No role/custom claim found for the authenticated user")

        return None

    return claims["role"]


def get_principal_credentials(
    user_email: str,
    client_oauth_secret: str,
    id_token: str,
    firebase_default_service_account: str,
) -> Tuple[Optional[service_account.Credentials], AuthenticationInfo.UserRole]:
    """Request the Service Account Credentials to Firestore using the current
    user logged Role.

    Args:
        user_email(str): The user email to request the claims for.
        client_oauth_secret(str): The OAuth 2.0 client ID for web application.
        id_token(str): The Identity Platform ID Token for the user.

    Returns:
        The service_account.Credentials for the user. None if there's an error.
    """
    firebase_app: firebase_admin.App = _get_firebase_default_app(
        firebase_default_service_account
    )
    if firebase_app is None:
        logging.error(
            "Could not create the default Firebase App. "
            "Principal Credentials could not be obtained"
        )
        return None, None

    user_role = _get_user_role(user_email)
    if not user_role:
        logging.error("No user role found for the user.")
        return None, None

    project_id = json.loads(client_oauth_secret)["web"]["project_id"]

    # Perform Request to Firestore.
    # TODO: This request should be more reliable.
    firestore_url = "https://firestore.googleapis.com/v1"
    default_db_path = f"projects/{project_id}/databases/(default)/documents"
    collection = f"user_roles/{user_role}"

    logging.info(f"Requesting service account for user role: {user_role}")

    headers = {
        "Authorization": "Bearer " + id_token,
        "Content-Type": "application/json",
    }
    response = requests.get(
        f"{firestore_url}/{default_db_path}/{collection}", headers=headers
    )

    if response.status_code == 200:
        try:
            content = json.loads(response.content.decode())
            service_account_info = json.loads(content["fields"]["key"]["stringValue"])
            role_credentials = service_account.Credentials.from_service_account_info(
                service_account_info
            )
            logging.info(
                f"SUCCESS obtaining service account credentials for "
                f"the user role: {user_role}"
            )
            return role_credentials, AuthenticationInfo.UserRole(user_role)

        except ValueError:
            logging.error(
                "Could not create service principal credentials for the "
                "service account found"
            )
            return None, None
        finally:
            firebase_admin.delete_app(firebase_admin.get_app())

    else:
        logging.error(
            f"Unexpected response code when requesting "
            f"service account for user role: {response.status_code}"
        )
        firebase_admin.delete_app(firebase_admin.get_app())
        return None, None



import os
if os.name=='nt':       # Windows
   import _winreg as winreg
elif os.name=='posix':  # Linux
   import glob

import serial
import time
import threading

from SmartMeshSDK.IpMoteConnector      import IpMoteConnector
from SmartMeshSDK.IpMgrConnectorSerial import IpMgrConnectorSerial
from SmartMeshSDK.ApiException         import ConnectionError

class SerialScanner(object):
    
    WAITFORMGRHELLO_TOUT = 3.500
    
    def __init__(self):
        pass
    
    # ======================= public ==========================================
    
    # === raw serial port manipulation
    
    def listAllSerialPorts(self):
        returnVal = []
        
        if os.name=='nt':
            key  = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                'HARDWARE\\DEVICEMAP\\SERIALCOMM'
            )
            for i in range(winreg.QueryInfoKey(key)[1]):
                try:
                    val = winreg.EnumValue(key,i)
                except:
                    pass
                else:
                    if val[0].find('VCP')>-1:
                        returnVal.append(str(val[1]))
        elif os.name=='posix':
            returnVal = glob.glob('/dev/ttyUSB*')
        
        returnVal.sort()
        
        return returnVal
    
    def canOpen(self,serialport):
        try:
            serialHandler = serial.Serial(serialport)
        except serial.serialutil.SerialException:
            returnVal = False
        else:
            returnVal = True
            serialHandler.close()
        return returnVal
    
    # === atomic testing of serial port
    
    def isIpManagerApi(self,serialport):
        return self._isIpApi(
            serialport       = serialport,
            apiClass         = IpMgrConnectorSerial.IpMgrConnectorSerial,
        )
    
    def isIpManagerCli(self,serialport):
        raise NotImplementedError()
    
    def isIpMoteApi(self,serialport):
        return self._isIpApi(
            serialport       = serialport,
            apiClass         = IpMoteConnector.IpMoteConnector,
            apiTestCommand   = 'dn_getParameter_moteInfo',
        )
    
    def isIpMoteCli(self,serialport):
        raise NotImplementedError()
    
    # === detecting IP manager's
    
    def listIpManagerApi(self):
        isManager  = {}
        dataLock   = threading.RLock()
        threads    = []
        for serialport in self.listAllSerialPorts():
            threads += [self._waitForMgrHello(serialport,isManager,dataLock)]
        for t in threads:
            t.join()
        return [k for (k,v) in isManager.items() if v==True]
    
    def availableManagerNotifier(self,cb,period=0):
        # start thread which continuously scans for IP Managers
        self._availableManagerNotifierThread(
            cb         = cb,
            period     = period,
        )
    
    # ======================= private =========================================
    
    def _isIpApi(self,serialport,apiClass,apiTestCommand=None):
        connector = apiClass()
        connector.MAX_NUM_RETRY = 1
        try:
            connector.connect({'port': serialport,})
            if apiTestCommand:
                getattr(connector,apiTestCommand)()
        except ConnectionError:
            returnVal = False
        else:
            returnVal = True
        finally:
            try:
                connector.disconnect()
            except:
                pass
            del connector
        return returnVal
    
    class _availableManagerNotifierThread(threading.Thread):
        def __init__(self,cb,period):
            self.cb         = cb
            self.period     = period
            threading.Thread.__init__(self)
            self.name       = "_ipManagerFinderThread"
            self.daemon     = True
            self.start()
        def run(self):
            while True:
                for m in SerialScanner().listIpManagerApi():
                    self.cb(m)
                time.sleep(self.period)

    class _waitForMgrHello(threading.Thread):
        def __init__(self,serialport,isManager,dataLock):
            self.serialport       = serialport
            self.isManager        = isManager
            self.dataLock         = dataLock
            self.goOn             = True
            with self.dataLock:
                isManager[serialport]=False
            threading.Thread.__init__(self)
            self.name             = "_waitForMgrHello@{0}".format(self.serialport)
            self.daemon           = True
            self.start()
        def run(self):
            # listen to that serial port for some time
            try:
                serialHandler = serial.Serial(self.serialport,baudrate=115200)
                serialHandler.setRTS(False)
                serialHandler.setDTR(True)
                listenThread = threading.Thread(
                    target = self._listenForMgrHello,
                    args = (
                        self.serialport,
                        serialHandler,
                        self.isManager,
                        self.dataLock,
                    ),
                )
                listenThread.name      = 'listenThread@{0}'.format(self.serialport)
                listenThread.daemon    = True
                listenThread.start()
                listenThread.join(SerialScanner.WAITFORMGRHELLO_TOUT)
                self.goOn = False
                serialHandler.close()
                while listenThread.isAlive():
                    pass # wait for listenThread to stop
            except serial.SerialException:
                pass # happens when serial port unavailable
        def _listenForMgrHello(self,serialport,serialHandler,isManager,dataLock):
            try:
                rxBuff = []
                while True:
                    c = serialHandler.read(1)
                    if c:
                        rxBuff += [ord(c)]
                        if self._contains_mrgHello(rxBuff):
                            with self.dataLock:
                                isManager[serialport]=True
                            return
            except Exception as err:
                if self.goOn:
                    print err
        def _contains_mrgHello(self,l):
            MSG_HELLO    = [126,0,3,0,2,4,0,155,56,126]
            len_mrgHello = len(MSG_HELLO)
            return any(l[i:len_mrgHello+i]==MSG_HELLO for i in xrange(len(l) - len_mrgHello+1))



"""
This module implements the MainAppComponent object which is the
object that contains the global state of the app and the main components:
- left sidebar
- Main View implementations
- Right Sidebar

The global state of the app is shared to the child components.
"""
from txp.cloud.controllers.authentication import AuthenticationInfo
from txp.web.core_components.app_component import AppComponent, AppState
from txp.web.core_components.sidebar import Sidebar
from txp.web.core_components.main_view import MainView
from txp.web.core_components.authentication import AuthenticationComponent
from txp.web.core_components.app_profiles import AppProfile
import streamlit as st
from typing import Dict
import logging
# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings
from txp.web.views import CommitView, EditConfigurationView, TrainingView, ManagementModelsView
log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class MainAppComponent(AppComponent):
    """The Main Application component. All the other components must be children of this component,
    directly or indirectly.

    - Any state with global context should be initialized here in this class:
        - Login information
        - Current selected profile
        - Current selected view

    This component has two direct kind of children:

        - The _sidebar_component, which is the left sidebar that acts as the "menu" of the app.

        - _main_active_components: Instances of the MainView that are available in the application,
            When new MainView's are added to the Application, a single instance of those views should be
            handled to this MainAppComponent at initialization time of the streamlit session for the app.
    """
    def __init__(self, default_profile: AppProfile = AppProfile.CONFIGURATION):
        super(MainAppComponent, self).__init__(
            component_key='app_layout'
        )
        # Children components
        self._sidebar_component: Sidebar = Sidebar()
        self._authentication_component: AuthenticationComponent = AuthenticationComponent()
        # A dict of MainView components, one of which will be rendered in the next streamlit run based
        # on the current state.
        self._main_active_components: Dict[str, MainView] = {}
        self._register_app_state(AppState())
        self._admin_views_created: bool = False

    def add_child_component(self, component: MainView, make_current: bool = False):
        """Adds the MainView component instance to the internal main active components.

        Args:
            component(MainView): A MainView component instance.
            make_current: True to set the current view to the passed component
        """
        if make_current:
            self.app_state.current_view = component.get_display_name()
            self.app_state.current_profile = component.get_associated_profile()
        self._main_active_components[component.get_display_name()] = component
        profile = component.get_associated_profile().get_display_value()
        self.app_state.views_for_profile[profile].append(component.get_display_name())

    def _render(self) -> None:
        if self.app_state.authentication_info.is_user_logged_in():

            if not self.app_state.project_model:
                logging.info("Project Model data not downloaded. Donwloading project data.")
                self._donwload_project_data()

            if self.app_state.authentication_info.user_role == AuthenticationInfo.UserRole.Admin \
                    and not self._admin_views_created:
                self.add_child_component(CommitView())
                self.add_child_component(ManagementModelsView())
                self.add_child_component(EditConfigurationView())
                self.add_child_component(TrainingView())
                self._admin_views_created = True

            self._sidebar_component.render()
            main_view_to_render = self._main_active_components.get(self.app_state.current_view, None)

            if (not main_view_to_render or main_view_to_render.get_display_name() not
                    in self.app_state.views_for_profile[self.app_state.current_profile.get_display_value()]):
                st.warning("Vista no disponible")

            else:
                main_view_to_render.render()

        else:
            self._authentication_component.render()


"""
This module offers MQTT utilities for the Gateway object.
"""
# ====================== Imports ===============================
from txp.common.config import settings
import paho.mqtt.client as paho_mqtt_client
from typing import Callable, Optional, List, Dict, Set, Union, Tuple
import ssl
import time
import threading
import jwt
import datetime
from txp.common.config import settings
import logging
import random
from txp.devices.exceptions import (
    MqttNetworkingUnavailableError,
    MqttPrivateKeyError,
    MqttConnectTimeoutError,
    MqttMessageNotPublished
)
import enum
from collections import defaultdict
import dataclasses

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ===================== Data ===================================
mqtt_bridge_hostname = settings.gateway.mqtt.mqtt_bridge_hostname
mqtt_bridge_port = settings.gateway.mqtt.mqtt_bridge_port
jwt_expires_hours = settings.gateway.mqtt.jwt_expires_hours
seconds_per_hour = 3600


class MQTTMessageType(enum.Enum):
    """Enumerated message types for the messages published in MQTT broker"""
    TELEMETRY_MESSAGE = 0
    ATTACH_MESSAGE = 1
    DETACH_MESSAGE = 2
    STATE_MESSAGE = 3
    DEVICES_PAIRING_MESSAGE = 4

    def __str__(self):
        return self.name


@dataclasses.dataclass
class MQTTSentPayloadContext:
    """This context allows to store unacknowledged messages in
    order to retry sending"""
    device_id: str
    topic: str
    payload: Union[str, bytes]
    attach: bool

    def __post_init__(self):
        # A timestamp to handle some reasonable timeout for
        # unacknowledged messages
        self.timestamp = time.time()


# ==============================================================================
# Definition of the GatewayConfigurationListener class.
# ==============================================================================
class GatewayMQTTClient:
    def __init__(
            self,
            gateway_ref: "Gateway",
            on_message_callback: Callable
    ):
        """Args:

            gateway_ref: the reference to the Gateway instance.
            on_message_callback: The callback to be executed by the client when a message
                is received. The Gateway should pass the instance method for this callback,
                and any processing will be performed by the Gateway instance.
        """
        self._project_id = gateway_ref.gateway_config.project_id
        self._cloud_region = gateway_ref.gateway_config.cloud_region
        self._registry_id = gateway_ref.gateway_config.registry_id
        self._gateway_id = gateway_ref.gateway_config.gateway_id
        self._on_message_callback = on_message_callback

        self._acknowledge_rtt_timeout = 10

        self._connected_flag = False
        self._disconnected_flag = True
        self._networking_fatal_error = False  # A networking fatal error makes it impossible for the MQTT library
        # to create a Client instance.

        self._client: paho_mqtt_client.Client = None

        # seconds since the token was issued
        # https://cloud.google.com/iot/docs/how-tos/credentials/jwts
        self._jwt_iat: int = 0

        # Control and state attributes
        self._subscribed_topics: Dict[str, str] = {}
        """_subscribed_topics: A dict to hold the subscribed topics and devices ID for when 
        the client is restarted"""

        self._attached_devices: Set[str] = set()
        """_attached_devices: To track attached devices between client restarts"""

        # The initial backoff time after a disconnection occurs, in seconds.
        self._MINIMUM_BACKOFF_TIME = 1
        self._MAXIMUM_BACKOFF_TIME = 64
        self._should_backoff: bool = False

        self._connection_resource_lock: threading.RLock = threading.RLock()

        self._tracked_messages: Dict[int, Tuple[bool, MQTTMessageType, Union[str, bytes, MQTTSentPayloadContext]]] = defaultdict(tuple)
        """_tracked_messages: 
        
        A Map of messages ids -> (bool, message type, payload). 
            The tuple elements are: 
                0 -> bool to indicate if the message was acknowledge or not by the broker.
                1 -> The MQTTMessageType provided when a client of the class publish a message.
                2 -> The payload published. 
                
        This internal table helps the client to track the result of
        published messages, specifically if they were delivered to the broker and all the 
        appropriate handshakes were made."""

        try:
            self._refresh_client()

        except MqttConnectTimeoutError:
            log.error("The Mqtt client could not establish a connection at initialization time")

        except MqttNetworkingUnavailableError:
            log.error("The MQTT client could not establish a connection at initialization time.")

    def _refresh_client(self) -> None:
        """Internally creates a new paho MQTT client and assigns the corresponding
        values in the instance attributes.

        Raises:
            MqttConnectionTimeoutError: if the client could not connect to the MQTT broker
                after the configured timeout.
            MqttNetworkingUnavailableError: if the client could not connect because there's an
                unknown error with the networking of the operative system.
        """
        try:
            if self._should_backoff:
                if self._MINIMUM_BACKOFF_TIME < self._MAXIMUM_BACKOFF_TIME:
                    self._MINIMUM_BACKOFF_TIME = self._MINIMUM_BACKOFF_TIME * 2
                    log.info(f'Increasing Exponential Backoff minimum to: {self._MINIMUM_BACKOFF_TIME} seconds')

            self._client: paho_mqtt_client.Client = self._get_client(
                self._project_id, self._cloud_region, self._registry_id, self._gateway_id,
                on_connect_callback=self._on_connect,
                on_disconnect_callback=self._on_disconnect,
                on_message=self._on_message_callback,
                should_backoff=self._should_backoff,
                backoff_time=self._MINIMUM_BACKOFF_TIME,
                on_publish_callback=self._on_publish_callback
            )

            start_wait = time.time()
            while not self._connected_flag:
                end_wait = time.time()
                self._client.loop(timeout=2)
                if (end_wait - start_wait) > settings.gateway.mqtt.connection_timeout_secs:
                    raise MqttConnectTimeoutError()

            self._jwt_iat = datetime.datetime.utcnow()
            log.info("New paho MQTT client connected")

            for attached_device in self._attached_devices:
                self._attach_device(attached_device, "")
                log.debug(f'Device re-attached for new client: {attached_device}')

            time.sleep(settings.gateway.mqtt.attach_timeout_secs)

            for topic in self._subscribed_topics:
                self._client.subscribe(topic, qos=1)
                log.debug(f'Topic resubscribed for new client: {topic}')

        except MqttConnectTimeoutError:
            log.error("Timeout exceeded to establish connection with MQTT Broker")
            self._connected_flag = False
            self._disconnected_flag = True
            self._should_backoff = True
            raise

        except Exception as e:
            log.error(f'Unexpected exception connecting the MQTT client with connect() method: {e}')
            self._connected_flag = False
            self._disconnected_flag = True
            self._should_backoff = True
            self._networking_fatal_error = True
            raise MqttNetworkingUnavailableError

    def enable_incoming_messages_processing(self, duration_secs: int):
        """Enables the MQTT client to process callbacks for incoming messages that are found
        in the incoming messages buffer.


        Raises:
            MqttConnectionTimeoutError: if the client tries to reconnect and fails.
            MqttNetworkingUnavailableError: if the client tries to reconnect and the networking presents
                an unknown problem.
        """
        try:
            self._connection_resource_lock.acquire()

            if self._networking_fatal_error or self._disconnected_flag:
                log.info('MQTT client disconnected. Trying to reconnect.')
                self._refresh_client()

            error_code = self._client.loop_start()

            if error_code is not None:  # Somebody already started the loop. Just return.
                log.debug("Other thread already called enable_incoming_messages_processing. "
                          " Leaving silently.")
                return

            time.sleep(duration_secs)

            self._client.loop_stop()

            if self._disconnected_flag:  # the incoming messages processed a disconnection. Try to reconnect immediately
                self._refresh_client()

            if self._jwt_expired():  # Tries to refresh if JWT expired
                self._refresh_jwt()

        except Exception as e:
            log.error(f'Exception encountered while enabling incoming messages: {e}.')
            raise

        finally:
            self._connection_resource_lock.release()

    def _jwt_expired(self) -> bool:
        """Returns True if the JSON Web Token expired."""
        seconds_since_issue = (datetime.datetime.utcnow() - self._jwt_iat).seconds
        return seconds_since_issue > seconds_per_hour * jwt_expires_hours

    def _refresh_jwt(self):
        """Creates a new internal paho MQTT client with a new token"""
        log.info('Refreshing JWT for MQTT client')
        self._client.loop_stop()
        self._client.disconnect()
        while not self._disconnected_flag:
            self.enable_incoming_messages_processing(1)  # Waits for disconnect processing

        self._refresh_client()

    def _on_connect(self, unused_client, unused_userdata, unused_flags, rc):
        """Callback to be executed asynchronously when the internal self._client
        gets connected."""
        log.info(f"Connection to MQTT IoT Bridge established")
        self._connected_flag = True
        self._disconnected_flag = False
        self._should_backoff = False
        self._networking_fatal_error = False
        self._MINIMUM_BACKOFF_TIME = 1

    def _on_disconnect(self, unused_client, unused_userdata, rc):
        """Callback to be executed asynchronously when the internal self._client
        is disconnected from the broker"""
        log.info(f"MQTT client disconnected. Disconnect Code:{rc}")
        if rc != 0:
            log.warning(f'Unexpected disconnect code in MQTT client: {rc}')
        self._disconnected_flag = True
        self._connected_flag = False
        self._should_backoff = True

    def _on_publish_callback(self, client, userdata, mid):
        """Callback to be executed asynchronously when a published message was acknowledge and
        it was delivered to the broker successfully."""
        if mid not in self._tracked_messages:
            log.debug(f"The message {mid} was delivered to the broker, "
                      f"but no tracking is active for that message id. Nothing is done.")

        else:
            value = self._tracked_messages[mid]
            self._tracked_messages[mid] = (True, value[1], value[2])
            log.debug(f"The message {mid} was delivered to the broker, "
                      f"Delivered tracking of the message is set to True. ")

    def subscribe_to_topic(self, device_id: str, topic: str, attach: bool) -> None:
        """Subscribe the client to the specified topic.

        Args:
            device_id: the device_id of the device related to the topic.
            topic: the topic to subscribe to.
            attach: True if the device must be 'attached' before subscribe. Edge devices must be 'attached' to the
                Gateway to allow the Gateway to publish telemetry in the name of the devices.

        Reference:
            https://github.com/GoogleCloudPlatform/python-docs-samples/blob/fca5081e785a19bcf28f2bde69240bf96b40d58f/iot/api-client/mqtt_example/cloudiot_mqtt_example.py#L210

        Returns:
            None
        """
        try:
            self._connection_resource_lock.acquire()

            if self._networking_fatal_error or self._disconnected_flag:
                log.info('MQTT client disconnected. Trying to reconnect.')
                self._refresh_client()

            if topic in self._subscribed_topics.keys():
                return

            if attach:
                self._attach_device(device_id, "")
                time.sleep(settings.gateway.mqtt.attach_timeout_secs)

            self._client.subscribe(topic, qos=1)
            self._subscribed_topics[topic] = device_id
            log.info(f'MQTT client subscribed to topic: {topic}')

        except Exception as e:
            log.error(f'Exception encountered while subscribing to topic: {e}')
            raise

        finally:
            self._connection_resource_lock.release()

    def publish_message(
            self,
            message_type: MQTTMessageType,
            device_id: str,
            topic: str,
            payload: Union[str, bytes],
            attach: bool,
            track_acknowledge: bool,
            try_reconnect: bool = True
    ):
        """Publish the message payload to the provided topic

        Published messages can be requested to be tracked by the MQTT client
        if the user of the class requires to know that a message was delivered or not.

        Args:
            message_type: The message type to be published.
            device_id: The device ID in IoT Core.
            topic: The MQTT Topic in the IoT Core Broker.
            payload: the message payload.
            attach: True if the device requires to be attached to the
                Gateway owner of this MQTT Client.
            track_acknowledge: If True, the MQTT client will track the
                published message on the on_publish() callback
                More info about on_publish() callback here:
                    https://www.eclipse.org/paho/index.php?page=clients/python/docs/index.php#callbacks

        """
        try:
            self._connection_resource_lock.acquire()

            if self._networking_fatal_error or self._disconnected_flag:
                if not try_reconnect:
                    log.error(f'There was an error publishing a message to the topic: {topic}')
                    raise MqttMessageNotPublished
                else:
                    log.info('MQTT client disconnected. Trying to reconnect.')
                    self._refresh_client()

            if self._jwt_expired():
                self._refresh_jwt()

            if attach and not self._is_attached(device_id):
                self._attach_device(device_id, "")
                time.sleep(settings.gateway.mqtt.attach_timeout_secs)

            (result, mid) = self._client.publish(topic, payload, qos=1)
            log.debug(f'Publishing message in topic: {topic}. '
                      f'mid: {mid}. result:{result}')

            if result == paho_mqtt_client.MQTT_ERR_SUCCESS:
                log.debug(f'Message successfully published for topic {topic}')
                if track_acknowledge:
                    if mid in self._tracked_messages:
                        log.warning(f"The message id {mid} was found in the tracked messages table."
                                    f" Overriding previous value.")
                    self._tracked_messages[mid] = (
                        False,
                        message_type,
                        MQTTSentPayloadContext(device_id, topic, payload, attach)
                    )
            else:
                log.error(f'There was an error publishing a message to the topic: {topic}')
                raise MqttMessageNotPublished

        except Exception as e:
            log.error(f'Unexpected Exception in MQTT connection operation: {e}')
            raise MqttMessageNotPublished

        finally:
            self._connection_resource_lock.release()

    def get_unacknowledged_messages(self, message_type: MQTTMessageType) -> List:
        """Returns a list of unsent messages according to the message type.

        The returned messages will be cleaned from the client tracking.

        Note: The list is not ordered according to how the messages were sent.

        Returns:
            List of unacknowledged payloads.
        """
        filtered_list = []
        items_to_pop = []
        log.debug(f"Checking for unacknowledged messages of type: {message_type.__str__()}")

        for mid, entry in self._tracked_messages.copy().items():
            is_sent, m_type, payload = entry
            if not is_sent and m_type == message_type:
                log.debug(f"Message with id {mid} was not acknowledge. Computing timeout valid")
                timeout_exceeded = (time.time() - payload.timestamp) > self._acknowledge_rtt_timeout
                if timeout_exceeded:
                    log.debug(f"The message {mid} exceeded timeout. Returning as unacknowledged")
                    filtered_list.append(payload)
                    self._tracked_messages.pop(mid)

        log.debug(f"{len(filtered_list)} unacknowledged messages found")
        return filtered_list

    def clear_acknowledged_messages(self, message_type: MQTTMessageType) -> None:
        """Clears the internal tracking of acknowledged messages of type message_type

        Returns:
            None
        """
        init_sz = len(self._tracked_messages)
        self._tracked_messages = {
            m_id: m_entry for m_id, m_entry in self._tracked_messages.items()
            if not m_entry[0] and not m_entry[1] == message_type
            # If not sent and message type is not message_type
        }
        after_sz = len(self._tracked_messages)
        log.debug(f"Removed {init_sz-after_sz} sent messages from tracking registry")

    def unsubscribe_from_topic(self, topic: str):
        """Unsubscribe the client from the specified topic.

        Args:
            The topic to unsubscribe from.

        Returns:
            None
        """
        try:
            self._connection_resource_lock.acquire()

            if self._networking_fatal_error or self._disconnected_flag:
                log.info('MQTT client disconnected. Trying to reconnect.')
                self._refresh_client()

            if topic not in self._subscribed_topics:
                log.warning(f'Trying to unsubscribe from topic {topic}, '
                            f'but the topic was not previously subscribed')
                return

            device_id = self._subscribed_topics[topic]

            if device_id in self._attached_devices:
                self._detach_device(device_id)

            self._client.unsubscribe(topic)
            log.info(f'MQTT client unsubscribed from topic: {topic}')

            self._subscribed_topics.pop(topic)

        except Exception as e:
            log.error(f'Unexpected Exception in MQTT connection operation: {e}')
            raise

        finally:
            self._connection_resource_lock.release()

    def _detach_device(self, device_id):
        """Detach the device from the gateway."""
        detach_topic = settings.gateway.mqtt.device_dettach_topic_template.format(
            device_id
        )
        log.debug("Detaching: {}".format(detach_topic))
        self._client.publish(detach_topic, "{}", qos=1)
        self._attached_devices.remove(device_id)

    def _attach_device(self, device_id, auth):
        """Attach the device to the gateway."""
        log.debug("Attaching: {}".format(device_id))
        attach_topic = settings.gateway.mqtt.device_attach_topic_template.format(
            device_id
        )
        attach_payload = '{{"authorization" : "{}"}}'.format(auth)
        self._client.publish(attach_topic, attach_payload, qos=1)
        self._attached_devices.add(device_id)

    def _is_attached(self, device_id: str) -> bool:
        """Returns True if the device is attached to the Gateway client."""
        return device_id in self._attached_devices

    @staticmethod
    def _create_jwt(project_id: str):
        """Creates a JWT to establish an MQTT connection

        Raises:
            MqttPrivateKeyError: If could not deserialize the Private Key data.
        """
        algorithm = "RS256"  # I created the key with this algorithm
        token = {
            "iat": datetime.datetime.utcnow(),
            "exp": datetime.datetime.utcnow() + datetime.timedelta(hours=jwt_expires_hours),
            "aud": project_id
        }

        try:
            with open(settings.gateway.private_key_path, "r") as f:
                private_key = f.read()

            log.info("Creating JWT using {} from private key file {}".format(
                algorithm, settings.gateway.private_key_path
            ))

            return jwt.encode(token, private_key, algorithm=algorithm)

        except:
            raise MqttPrivateKeyError()

    @classmethod
    def _get_client(
            cls,
            project_id: str,
            cloud_region: str,
            registry_id: str,
            device_id: str,
            on_connect_callback: Optional[Callable] = None,
            on_disconnect_callback: Optional[Callable] = None,
            on_message: Optional[Callable] = None,
            on_publish_callback: Optional[Callable] = None,
            should_backoff: bool = False,
            backoff_time: int = 0
    ):
        """Creates a Paho mqtt Client instance to handle the MQTT protocol communication"""
        client_id = settings.gateway.mqtt.device_id_template.format(
            project_id, cloud_region, registry_id, device_id
        )
        log.info(f"MQTT client ID configured {client_id}")

        client = paho_mqtt_client.Client(client_id=client_id)

        client.username_pw_set(
            username='unused',
            password=cls._create_jwt(project_id)
        )

        client.tls_set(
            ca_certs=settings.gateway.ca_certs,
            tls_version=ssl.PROTOCOL_TLSv1_2
        )

        client.on_connect = on_connect_callback
        client.on_disconnect = on_disconnect_callback
        client.on_message = on_message
        client.on_publish = on_publish_callback

        if should_backoff:
            delay = backoff_time + random.randint(0, 1000) / 1000.0
            time.sleep(delay)

        client.connect(mqtt_bridge_hostname, mqtt_bridge_port,
                       settings.gateway.mqtt.keep_alive_minutes * 60)

        return client

    @staticmethod
    def validate_connection_values(
        project_id: str,
        cloud_region: str,
        registry_id: str,
        device_id: str,
    ) -> bool:
        """Validate if a Cloud Connection could be established using the passed
            arguments and the existing private key.

            Args:
                project_id: the project ID in Cloud IoT
                cloud_region: the cloud region configured for the project in Cloud IoT
                registry_id: the registry ID configured in the project
                device_id: the device ID of the Gateway (the serial)

            Raises:
                ValueError: If could not deserialize the Private Key data.

            Returns:
                True if the connection was established successfully. False otherwise.
        """

        def on_connect_closure(flags):
            """Callback for when a device connects."""

            def on_connect_callback(unused_client, unused_userdata, unused_flags, rc):
                log.info(f"Connection to MQTT IoT Bridge established")
                flags[0] = True

            return on_connect_callback

        def on_disconnect(unused_client, unused_userdata, rc):
            """Paho callback for when a device disconnects."""
            message = paho_mqtt_client.connack_string(rc)
            log.info(f"Disconnected from MQTT IoT Bridge {message}")

        def on_message_closure(flags):
            def on_message_callback(unused_client, unused_userdata, message):
                """Callback when the device receives a message on a subscription."""
                flags[1] = True
                log.info(f"Message received from MQTT IoT Bridge received")

            return on_message_callback

        flags = [False, False]

        try:
            client: paho_mqtt_client.Client = GatewayMQTTClient._get_client(
                project_id, cloud_region, registry_id, device_id,
                on_connect_closure(flags), on_disconnect, on_message_closure(flags)
            )
        except ValueError as key_deseriaize_error:
            raise key_deseriaize_error

        client.connect(mqtt_bridge_hostname, mqtt_bridge_port)

        # Subscribe to the Gateway topic.
        mqtt_config_topic = "/devices/{}/config".format(device_id)

        client.subscribe(mqtt_config_topic, qos=1)

        jwt_iat = datetime.datetime.utcnow()
        test_start = time.time()

        while True:
            client.loop_start()

            seconds_since_issue = (datetime.datetime.utcnow() - jwt_iat).seconds
            if seconds_since_issue > seconds_per_hour * jwt_expires_hours:
                log.info("Refreshing token after {}s".format(seconds_since_issue))
                jwt_iat = datetime.datetime.utcnow()
                client.loop_stop()
                client.disconnect()
                client = GatewayMQTTClient._get_client(
                    project_id, cloud_region, registry_id, device_id,
                    on_connect_closure(flags), on_disconnect, on_message_closure(flags)
                )

            client.loop_stop()
            time.sleep(1)  # sleeps for 1 second for next loop activation

            elapsed = time.time() - test_start

            if elapsed > 15 or all(flags):
                client.disconnect()
                break

        if all(flags):
            log.info(f"Connection to MQTT IoT Bridge validated")
            return True
        else:
            log.warning(f"Could not validate Connection to MQTT IoT Bridge")
            return False



import numpy as np
from common import iCOMOX_messages, messages_utils
from scipy.signal import welch
from scipy.signal.windows import hamming
import struct


class class_ADXL356_SMIP:
    def __init__(self):
        # Initializing Fs_Acc, MFFT_Acc, window
        Root_clk = 26e6
        ACLKDIV = 2  # ADC clock is determined by Root_clk/ACLKDIV
        SAMPLETIME = 255    #116  # AcquisitionTime-1 in terms of ADC clocks
        OverSMP = 1  # No over sample, No averaging in digital domain
        NumOfCh = 1  # Number of interleaved ChannelsBitmask
        DLY = 0  # Delay at the end of channels round, in units of ADC clocks

        if (DLY == 0):
            self.Fs_Acc = (Root_clk / ACLKDIV) / ((14 + SAMPLETIME) * OverSMP * NumOfCh)
        else:
            if NumOfCh == 1:
                self.Fs_Acc = (Root_clk / ACLKDIV) / ((14 + SAMPLETIME) * OverSMP * NumOfCh + DLY + 2)
            else:
                self.Fs_Acc = (Root_clk / ACLKDIV) / ((14 + SAMPLETIME) * OverSMP * NumOfCh + DLY + 3)
        self.NFFT_Acc = iCOMOX_messages.ADXL356_SMIP_SAMPLES_NUM_PER_PIN    # 512
        self.Sensitivity_Acc = 80E-3  # 80mV/g for range of +/-10g. 40mV/g for range of +/-20g. 0g is translated to 900mV
        self.window = hamming(M=self.NFFT_Acc)

    def ADC_readings_to_g_units(self, ADC_samples):
        # convert ADC samples to g units
        return np.array(ADC_samples) * 1.8 / (4096 * self.Sensitivity_Acc)

    def g_units_to_PSD(self, acc_units_g):
        (freq_Acc, Pxxf_Acc) = welch(x=acc_units_g, fs=self.Fs_Acc, window=self.window, noverlap=self.NFFT_Acc/2, nfft=self.NFFT_Acc, return_onesided=True, detrend=False)
        return freq_Acc, Pxxf_Acc

    def ADXL356_UnpackBytes(self, packed_bytes):  # arranging the compressed 12 bits in 16 bits elements (4 MSB bits of each element are 0)
        unpacked_bytes = bytearray(iCOMOX_messages.ADXL356_SMIP_SAMPLES_NUM_PER_PIN * 2)
        src_index = 0  # pointer to compress_samples
        dst_index = 0  # pointer to unpacked_bytes
        for i in range(0, iCOMOX_messages.ADXL356_SMIP_SAMPLES_NUM_PER_PIN // 2):  # iterate on the total number of samples divided by 2
            unpacked_bytes[dst_index] = packed_bytes[src_index]
            unpacked_bytes[dst_index + 1] = packed_bytes[src_index + 1] & 0x0F
            unpacked_bytes[dst_index + 2] = (packed_bytes[src_index + 1] >> 4) + ((packed_bytes[src_index + 2] & 0x0F) << 4)
            unpacked_bytes[dst_index + 3] = packed_bytes[src_index + 2] >> 4
            dst_index += 4
            src_index += 3
        return unpacked_bytes

    def msg_bytes_to_samples_g(self, payload):
        unpacked_bytes = messages_utils.ADC_Unpack_12bits_Array(packed_bytes=payload)
        ADC_samples = np.array(struct.unpack("<{}h".format(iCOMOX_messages.ADXL356_SMIP_SAMPLES_NUM_PER_PIN), unpacked_bytes)) - 2048
        return self.ADC_readings_to_g_units(ADC_samples=ADC_samples)
